import ctypes

from .. import builder
from .. import consts


class Fr(ctypes.Structure):
    _fields_ = [("v", ctypes.c_ulonglong * consts.FR_SIZE)]


Fr.__add__ = builder.buildThreeOp(Fr, "add")
Fr.__eq__ = builder.buildIsEqual(Fr)
Fr.__mul__ = builder.buildThreeOp(Fr, "mul")
Fr.__neg__ = builder.buildTwoOp(Fr, "neg")
Fr.__sub__ = builder.buildThreeOp(Fr, "sub")
Fr.__truediv__ = builder.buildThreeOp(Fr, "div")
Fr.getStr = builder.buildGetStr(Fr)
Fr.isOne = builder.buildIsOne(Fr)
Fr.isZero = builder.buildIsZero(Fr)
Fr.serialize = builder.buildSerialize(Fr)
Fr.setByCSPRNG = builder.buildSetByCSPRNG(Fr)
Fr.setInt = builder.buildSetInt(Fr)
Fr.setStr = builder.buildSetStr(Fr)
Fr.setHashOf = builder.buildSetHashOf(Fr)
Fr.clear = builder.buildClear(Fr)



from eigensdk.crypto.bls.attestation import (
    G1Point,
    G2Point,
    GTPoint,
    Signature,
    PrivateKey,
    KeyPair,
    new_g1_point,
    new_g2_point,
    new_zero_g1_point,
    new_zero_g2_point,
    new_zero_signature,
    new_private_key,
    new_key_pair,
    new_key_pair_from_string,
    gen_random_bls_keys,
    g1_to_tupple,
    g2_to_tupple,
)



import ctypes

from .Fp import Fp
from .. import builder


class Fp2(ctypes.Structure):
    _fields_ = [("d", Fp * 2)]


Fp2.get_a = builder.buildGetSubArray(Fp2, 1, Fp)
Fp2.get_b = builder.buildGetSubArray(Fp2, 0, Fp)
Fp2.clear = builder.buildClear(Fp2)



import ctypes

from .. import builder
from .. import consts


class Fp(ctypes.Structure):
    _fields_ = [("v", ctypes.c_ulonglong * consts.FP_SIZE)]


Fp.__add__ = builder.buildThreeOp(Fp, "add")
Fp.__eq__ = builder.buildIsEqual(Fp)
Fp.__invert__ = builder.buildTwoOp(Fp, "inv")
Fp.__mul__ = builder.buildThreeOp(Fp, "mul")
Fp.__neg__ = builder.buildTwoOp(Fp, "neg")
Fp.__sub__ = builder.buildThreeOp(Fp, "sub")
Fp.__truediv__ = builder.buildThreeOp(Fp, "div")
Fp.deserialize = builder.buildDeserialize(Fp)
Fp.getStr = builder.buildGetStr(Fp)
Fp.isOne = builder.buildIsOne(Fp)
Fp.isZero = builder.buildIsZero(Fp)
Fp.serialize = builder.buildSerialize(Fp)
Fp.setByCSPRNG = builder.buildSetByCSPRNG(Fp)
Fp.setInt = builder.buildSetInt(Fp)
Fp.setStr = builder.buildSetStr(Fp)
Fp.clear = builder.buildClear(Fp)



import logging
from enum import IntEnum
from typing import List, Any, Dict

from eth_account.signers.local import LocalAccount
from web3 import Web3
from web3.contract import Contract
from web3.contract.contract import ContractFunction
from web3.types import ChecksumAddress
from web3.types import TxReceipt

from eigensdk.chainio.utils import (
    abi_encode_normal_registration_params,
    get_pubkey_registration_params,
    Transactor,
)
from eigensdk.types_ import Operator


class RegistrationType(IntEnum):
    NORMAL = 0
    TRUSTED = 1


class ELWriter:
    def __init__(
        self,
        allocation_manager: Contract,
        avs_directory: Contract,
        delegation_manager: Contract,
        permission_controller: Contract,
        reward_coordinator: Contract,
        registry_coordinator: Contract,
        strategy_manager: Contract,
        el_chain_reader: Any,
        eth_http_client: Web3,
        logger: logging.Logger,
        pk_wallet: LocalAccount,
        strategy_abi: List[Dict[str, Any]],
        erc20_abi: List[Dict[str, Any]],
    ):
        self.allocation_manager = allocation_manager
        self.avs_directory = avs_directory
        self.delegation_manager = delegation_manager
        self.permission_controller = permission_controller
        self.rewards_coordinator = reward_coordinator
        self.registry_coordinator = registry_coordinator
        self.strategy_manager = strategy_manager
        self.logger = logger
        self.strategy_abi = strategy_abi
        self.erc20_abi = erc20_abi
        self.el_chain_reader = el_chain_reader
        self.transactor = Transactor(pk_wallet, eth_http_client)

        if allocation_manager is None:
            self.logger.warning("AllocationManager contract not provided")

        if avs_directory is None:
            self.logger.warning("AvsDirectory contract not provided")

        if delegation_manager is None:
            self.logger.warning("DelegationManager contract not provided")

        if permission_controller is None:
            self.logger.warning("PermissionController contract not provided")

        if reward_coordinator is None:
            self.logger.warning("RewardCoordinator contract not provided")

        if registry_coordinator is None:
            self.logger.warning("RegistryCoordinator contract not provided")

        if strategy_manager is None:
            self.logger.warning("StrategyManager contract not provided")

        if el_chain_reader is None:
            self.logger.warning("ELChainReader contract not provided")

        if eth_http_client is None:
            self.logger.warning("EthHTTPClient not provided")

        if pk_wallet is None:
            self.logger.warning("PKWallet not provided")

        if strategy_abi is None:
            self.logger.warning("StrategyABI not provided")

        if erc20_abi is None:
            self.logger.warning("ERC20ABI not provided")

    def send_transaction(self, func: ContractFunction):
        return self.transactor.send(func)

    def register_as_operator(self, operator: Operator) -> TxReceipt:
        func = self.delegation_manager.functions.registerAsOperator(
            operator.delegation_approver_address,
            operator.allocation_delay,
            operator.metadata_url,
        )
        return self.send_transaction(func)

    def update_operator_details(self, operator: Operator) -> TxReceipt:
        func = self.delegation_manager.functions.modifyOperatorDetails(
            operator.address,
            operator.delegation_approver_address,
        )
        return self.send_transaction(func)

    def update_metadata_uri(self, operator_address: str, uri: str) -> TxReceipt:
        func = self.delegation_manager.functions.updateOperatorMetadataURI(
            Web3.to_checksum_address(operator_address),
            uri,
        )
        return self.send_transaction(func)

    def deposit_erc20_into_strategy(self, strategy_addr: str, amount: int) -> TxReceipt:
        _, token_contract, token_addr = (
            self.el_chain_reader.get_strategy_and_underlying_erc20_token(strategy_addr)
        )
        func = token_contract.functions.approve(
            self.strategy_manager.address,
            amount,
        )
        self.send_transaction(func)
        func = self.strategy_manager.functions.depositIntoStrategy(
            Web3.to_checksum_address(strategy_addr),
            Web3.to_checksum_address(token_addr),
            amount,
        )
        return self.send_transaction(func)

    def set_claimer_for(self, claimer: str) -> TxReceipt:
        func = self.rewards_coordinator.functions.setClaimerFor(
            Web3.to_checksum_address(claimer),
        )
        return self.send_transaction(func)

    def process_claim(self, claim: dict, recipient_address: str) -> TxReceipt:
        claim_tuple = (
            claim["rootIndex"],
            claim["earnerIndex"],
            claim["earnerTreeProof"],
            (  # earnerLeaf (tuple)
                Web3.to_checksum_address(claim["earnerLeaf"]["earner"]),
                claim["earnerLeaf"]["earnerTokenRoot"],
            ),
            claim["tokenIndices"],  # list of uint32
            claim["tokenTreeProofs"],  # list of bytes
            [
                (Web3.to_checksum_address(tl["token"]), int(tl["cumulativeEarnings"]))
                for tl in claim["tokenLeaves"]
            ],
        )

        func = self.rewards_coordinator.functions.processClaim(
            claim_tuple, Web3.to_checksum_address(recipient_address)
        )

        return self.send_transaction(func)

    def set_operator_avs_split(self, operator: str, avs: str, split: int) -> TxReceipt:

        func = self.rewards_coordinator.functions.setOperatorAVSSplit(
            Web3.to_checksum_address(operator),
            Web3.to_checksum_address(avs),
            split,
        )

        return self.send_transaction(func)

    def set_operator_pi_split(self, operator: str, split: int) -> TxReceipt:

        func = self.rewards_coordinator.functions.setOperatorPISplit(
            Web3.to_checksum_address(operator),
            split,
        )

        return self.send_transaction(func)

    def modify_allocations(
        self,
        operator_address: str,
        avs_service_manager: str,
        operator_set_id: int,
        strategies: list,
        new_magnitudes: list,
    ) -> TxReceipt:
        allocation = (
            (Web3.to_checksum_address(avs_service_manager), operator_set_id),
            [Web3.to_checksum_address(s) for s in strategies],
            new_magnitudes,
        )

        func = self.allocation_manager.functions.modifyAllocations(
            Web3.to_checksum_address(operator_address), [allocation]
        )

        return self.send_transaction(func)

    def clear_deallocation_queue(
        self, operator_address: str, strategies: list, nums_to_clear: list
    ) -> TxReceipt:

        func = self.allocation_manager.functions.clearDeallocationQueue(
            Web3.to_checksum_address(operator_address),
            [Web3.to_checksum_address(s) for s in strategies],
            nums_to_clear,
        )

        return self.send_transaction(func)

    def set_allocation_delay(self, operator_address: str, delay: int) -> TxReceipt:

        func = self.allocation_manager.functions.setAllocationDelay(
            Web3.to_checksum_address(operator_address),
            delay,
        )

        return self.send_transaction(func)

    def deregister_from_operator_sets(self, operator: str, request: dict) -> TxReceipt:

        func = self.allocation_manager.functions.deregisterFromOperatorSets(
            {
                "operator": Web3.to_checksum_address(operator),
                "avs": Web3.to_checksum_address(request["avs_address"]),
                "operatorSetIds": request["operator_set_ids"],
            }
        )

        return self.send_transaction(func)

    def register_for_operator_sets(
        self, registry_coordinator_addr: str, request: dict
    ) -> TxReceipt:
        pubkey_reg_params = get_pubkey_registration_params(
            self.transactor.eth_http_client,
            Web3.to_checksum_address(registry_coordinator_addr),
            Web3.to_checksum_address(request["operator_address"]),
            request["bls_key_pair"],
        )

        encoded_data = abi_encode_normal_registration_params(
            RegistrationType.NORMAL,
            request["socket"],
            pubkey_reg_params,
        )

        register_params = {
            "avs": Web3.to_checksum_address(request["avs_address"]),
            "operatorSetIds": request["operator_set_ids"],
            "data": encoded_data,
        }

        func = self.allocation_manager.functions.registerForOperatorSets(
            Web3.to_checksum_address(request["operator_address"]),
            register_params,
        )

        return self.send_transaction(func)

    def remove_permission(self, request: dict) -> TxReceipt:

        func = self.permission_controller.functions.removeAppointee(
            Web3.to_checksum_address(request["account_address"]),
            Web3.to_checksum_address(request["appointee_address"]),
            request["target"],
            request["selector"],
        )

        return self.send_transaction(func)

    def set_permission(self, request: dict) -> TxReceipt:

        func = self.permission_controller.functions.setAppointee(
            Web3.to_checksum_address(request["account_address"].lower()),
            Web3.to_checksum_address(request["appointee_address"].lower()),
            Web3.to_checksum_address(request["target"].lower()),
            request["selector"],
        )
        return self.send_transaction(func)

    def accept_admin(self, request: dict) -> TxReceipt:

        func = self.permission_controller.functions.acceptAdmin(
            Web3.to_checksum_address(request["account_address"]),
        )

        return self.send_transaction(func)

    def add_pending_admin(self, request: dict) -> TxReceipt:

        func = self.permission_controller.functions.addPendingAdmin(
            Web3.to_checksum_address(request["account_address"]),
            Web3.to_checksum_address(request["admin_address"]),
        )

        return self.send_transaction(func)

    def remove_admin(self, request: dict) -> TxReceipt:
        func = self.permission_controller.functions.removeAdmin(
            Web3.to_checksum_address(request["account_address"]),
            Web3.to_checksum_address(request["admin_address"]),
        )
        return self.send_transaction(func)

    def remove_pending_admin(self, request: dict) -> TxReceipt:

        func = self.permission_controller.functions.removePendingAdmin(
            Web3.to_checksum_address(request["account_address"]),
            Web3.to_checksum_address(request["admin_address"]),
        )

        return self.send_transaction(func)

    def get_operator_id(self, operator_address: ChecksumAddress) -> bytes:
        return self.registry_coordinator.functions.getOperatorId(operator_address).call()

    def set_avs_registrar(self, avs_address: str, registrar_address: str) -> TxReceipt:
        func = self.allocation_manager.functions.setAVSRegistrar(
            Web3.to_checksum_address(avs_address),
            Web3.to_checksum_address(registrar_address),
        )
        return self.send_transaction(func)



from web3 import Web3
from tests.builder import clients, clients_array, config


def advance_chain_by_n_blocks(web3_client, n: int):
    for _ in range(n):
        web3_client.provider.make_request("evm_mine", [])


def test_deregister_from_operator_sets():
    # Now deregister from the same operator sets
    operator_address = config["operator_address_1"]
    request = {
        "avs_address": config["service_manager_address"],
        "operator_set_ids": [0],
    }

    receipt = clients.el_writer.deregister_from_operator_sets(operator_address, request)

    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Deregistered from operator sets with tx hash: {receipt['transactionHash'].hex()}")


def test_update_metadata_uri():
    operator_addr = config["operator_address_1"]
    metadata_uri = "https://example.com/updated-metadata-uri"
    receipt = clients.el_writer.update_metadata_uri(operator_addr, metadata_uri)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Updated metadata URI with tx hash: {receipt['transactionHash'].hex()}")


def test_deposit_erc20_into_strategy():
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    amount = 100
    receipt = clients.el_writer.deposit_erc20_into_strategy(strategy_addr, amount)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Deposited ERC20 tokens into strategy with tx hash: {receipt['transactionHash'].hex()}")


def test_set_claimer_for():
    claimer_addr = config["operator_address_1"]
    receipt = clients.el_writer.set_claimer_for(claimer_addr)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set claimer with tx hash: {receipt['transactionHash'].hex()}")


def test_set_operator_avs_split():
    operator_addr = config["operator_address_1"]
    avs_addr = config["avs_address"]
    split = 5000
    receipt = clients.el_writer.set_operator_avs_split(operator_addr, avs_addr, split)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set operator AVS split with tx hash: {receipt['transactionHash'].hex()}")


def test_set_operator_pi_split():
    operator_addr = config["operator_address_1"]
    split = 3000
    receipt = clients.el_writer.set_operator_pi_split(operator_addr, split)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set operator PI split with tx hash: {receipt['transactionHash'].hex()}")


def test_clear_deallocation_queue():
    operator_addr = config["operator_address_1"]
    strategy_addr = config["strategy_addr"]
    strategies = [strategy_addr]
    nums_to_clear = [1]
    receipt = clients.el_writer.clear_deallocation_queue(operator_addr, strategies, nums_to_clear)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Cleared deallocation queue with tx hash: {receipt['transactionHash'].hex()}")


def test_set_allocation_delay():
    operator_addr = config["operator_address_1"]
    delay = 50
    receipt = clients.el_writer.set_allocation_delay(operator_addr, delay)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set allocation delay with tx hash: {receipt['transactionHash'].hex()}")


def test_add_pending_admin():
    request = {
        "account_address": config["operator_address_1"],
        "admin_address": config["operator_address_1"],
    }
    receipt = clients.el_writer.add_pending_admin(request)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Added pending admin with tx hash: {receipt['transactionHash'].hex()}")


def test_remove_pending_admin():
    request = {
        "account_address": config["operator_address_1"],
        "admin_address": config["operator_address_1"],
    }
    receipt = clients.el_writer.remove_pending_admin(request)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Removed pending admin with tx hash: {receipt['transactionHash'].hex()}")


def test_accept_admin():
    test_add_pending_admin()
    request = {"account_address": config["operator_address_1"]}
    receipt = clients.el_writer.accept_admin(request)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Accepted admin with tx hash: {receipt['transactionHash'].hex()}")


def test_set_permission():
    request = {
        "account_address": config["operator_address_1"],
        "appointee_address": config["operator_address_1"],
        "target": config["avs_address"],
        "selector": "0x12345678",
    }
    receipt = clients.el_writer.set_permission(request)
    assert receipt["status"] == 1


def test_remove_permission():
    request = {
        "account_address": config["operator_address_1"],
        "appointee_address": config["operator_address_1"],
        "target": config["avs_address"],
        "selector": "0x12345678",
    }
    receipt = clients.el_writer.remove_permission(request)
    assert receipt["status"] == 1


def test_remove_admin_flow():
    account_address = Web3.to_checksum_address(config["operator_address_1"])
    admin2_address = Web3.to_checksum_address(config["operator_address_2"])

    receipt = clients.el_writer.add_pending_admin(
        {
            "account_address": account_address,
            "admin_address": admin2_address,
            "wait_for_receipt": True,
        }
    )
    assert receipt["status"] == 1, f"Transaction failed: {receipt}"

    receipt = clients_array[1].el_writer.accept_admin(
        {
            "account_address": account_address,
        }
    )
    assert receipt["status"] == 1, f"Transaction failed: {receipt}"

    receipt = clients.el_writer.remove_admin(
        {
            "account_address": account_address,
            "admin_address": admin2_address,
        }
    )
    assert receipt["status"] == 1, f"Transaction failed: {receipt}"


def test_modify_allocations():
    operator_address = config["operator_address_1"]
    avs_service_manager = config["service_manager_address"]
    operator_set_id = 0
    strategies = [config["strategy_addr"]]
    new_magnitudes = [1000]
    receipt = clients.el_writer.modify_allocations(
        operator_address, avs_service_manager, operator_set_id, strategies, new_magnitudes
    )
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Modified allocations with tx hash: {receipt['transactionHash'].hex()}")



from .structures.Fr import Fr
from .structures.G1 import G1
from .structures.G2 import G2
from .structures.GT import GT
from .structures.Fp import Fp

__all__ = (
    "Fp",
    "Fr",
    "G1",
    "G2",
    "GT",
)



from mcl import G1, G2, GT, Fr

# modulus for the underlying field F_p of the elliptic curve
_FP_MODULUS = 21888242871839275222246405745257275088696311157297823662689037894645226208583

_G2_XA = 0x198E9393920D483A7260BFB731FB5D25F1AA493335A9E71297E485B7AEF312C2
_G2_XB = 0x1800DEEF121F1E76426A00665E5C4479674322D4F75EDADD46DEBD5CD992F6ED
_G2_YA = 0x090689D0585FF075EC9E99AD690C3395BC4B313370B38EF355ACDADCD122975B
_G2_YB = 0x12C85EA5DB8C6DEB4AAB71808DCB408FE3D1E7690C43D37B4CE6CC0166FA7DAA


def __addmod(a, b, m):
    return (a + b) % m


def __mulmod(a, b, m):
    return (a * b) % m


def __expmod(a, b, m):
    result = 1
    base = a
    _b = b
    while _b > 0:
        # Check the least significant bit (LSB) of b
        if _b & 1:
            result = (result * base) % m
        # Right shift b by 1 (effectively dividing by 2, discarding the remainder)
        _b >>= 1
        # Square the base for the next iteration (efficient for repeated multiplication)
        base = (base * base) % m
    return result


def __g1_point(x: int, y: int) -> G1:
    res = G1()
    res.setStr(f"1 {x} {y}".encode("utf-8"))
    return res


def verify_sig(
    sig: G1,
    pub_key: G2,
    msg_bytes: bytes,
) -> bool:
    G2 = get_g2_generator()
    msg_point = map_to_curve(msg_bytes)

    gt1 = GT.pairing(msg_point, pub_key)
    gt2 = GT.pairing(sig, G2)

    return gt1 == gt2


def map_to_curve(_x: bytes) -> G1:
    beta = 0
    y = 0
    x = int.from_bytes(_x, "big") % _FP_MODULUS
    while True:
        (beta, y) = __find_y_from_x(x)
        # y^2 == beta
        if beta == ((y * y) % _FP_MODULUS):
            return __g1_point(x, y)
        x = (x + 1) % _FP_MODULUS
    return __g1_point(0, 0)


def __find_y_from_x(x: int) -> "tuple[int, int]":
    # beta = (x^3 + b) % p
    beta = __addmod(__mulmod(__mulmod(x, x, _FP_MODULUS), x, _FP_MODULUS), 3, _FP_MODULUS)
    # y^2 = x^3 + b
    # this acts like: y = sqrt(beta) = beta^((p+1) / 4)
    y = __expmod(
        beta,
        0xC19139CB84C680A6E14116DA060561765E05AA45A1C72A34F082305B61F3F52,
        _FP_MODULUS,
    )
    return (beta, y)


def check_g1_and_g2_discrete_log_equality(p1: G1, p2: G2) -> bool:
    G1 = get_g1_generator()
    G2 = get_g2_generator()

    gt1 = GT.pairing(p1, G2)
    gt2 = GT.pairing(G1, p2)
    return gt1 == gt2


def get_g1_generator() -> G1:
    g1 = G1()
    g1.setStr(b"1 1 2")
    return g1


def get_g2_generator() -> G2:
    g2 = G2()
    g2.setStr(f"1 {_G2_XB} {_G2_XA} {_G2_YB} {_G2_YA}".encode("utf-8"))
    return g2


def mul_by_generator_g1(a: Fr) -> G1:
    return get_g1_generator() * a


def mul_by_generator_g2(a: Fr) -> G2:
    return get_g2_generator() * a



import json
import os
import secrets
from eth_account import Account
from mcl import G1, G2, GT, Fr, Fp
from eigensdk.crypto.bn256 import utils as bn256Utils


def new_fp_element(v: int) -> Fp:
    fp = Fp()
    fp.setInt(v)
    return fp


class G1Point(G1):
    def __init__(self, x: int, y: int) -> None:
        super().__init__()
        self.setStr(f"1 {x} {y}".encode("utf-8"))
        if x == 0 and y == 0:
            self.clear()

    def from_G1(g1: G1):
        x = int(g1.getX().getStr())
        y = int(g1.getY().getStr())
        return G1Point(x, y)

    def __add__(self, a: "G1Point"):
        return G1Point.from_G1(super().__add__(a).normalize())

    def __sub__(self, a: "G1Point"):
        return G1Point.from_G1(super().__sub__(a).normalize())

    def verify_equivalence(self, a: "G2Point"):
        return bn256Utils.check_g1_and_g2_discrete_log_equality(self, a)


def new_g1_point(x: int, y: int) -> G1Point:
    res = G1Point(x, y)
    if x == 0 and y == 0:
        res.clear()
    return res


def new_zero_g1_point() -> G1Point:
    return new_g1_point(0, 0)


class G2Point(G2):
    def __init__(self, xa: int, xb: int, ya: int, yb: int) -> None:
        super().__init__()
        self.setStr(f"1 {xb} {xa} {yb} {ya}".encode("utf-8"))
        if xa == 0 and xb == 0 and ya == 0 and yb == 0:
            self.clear()

    def from_G2(g2: G2):
        xa = int(g2.getX().get_a().getStr())
        xb = int(g2.getX().get_b().getStr())
        ya = int(g2.getY().get_a().getStr())
        yb = int(g2.getY().get_b().getStr())
        return G2Point(xa, xb, ya, yb)

    def __add__(self, a: "G2Point"):
        return G2Point.from_G2(super().__add__(a).normalize())

    def __sub__(self, a: "G2Point"):
        return G2Point.from_G2(super().__sub__(a).normalize())


def new_g2_point(xa: int, xb: int, ya: int, yb: int) -> G2Point:
    return G2Point(xa, xb, ya, yb)


def new_zero_g2_point() -> G2Point:
    return new_g2_point(0, 0, 0, 0)


class GTPoint(GT):
    pass


class Signature(G1Point):
    @staticmethod
    def from_g1_point(p: G1Point) -> "Signature":
        x = int(p.getX().getStr())
        y = int(p.getY().getStr())
        return Signature(x, y)

    def to_json(self) -> dict:
        return {
            "X": int(self.getX().getStr()),
            "Y": int(self.getY().getStr()),
        }

    def add(self, a: "Signature"):
        return self + a

    def verify(self, pub_key: G2Point, msg_bytes: bytes) -> bool:
        return bn256Utils.verify_sig(self, pub_key, msg_bytes)


def new_zero_signature() -> Signature:
    return Signature(0, 0)


class PrivateKey(Fr):
    def __init__(self, secret: bytes = None):
        super().__init__()
        if not secret:
            self.setHashOf(secrets.token_bytes(32))
        else:
            int_key = int.from_bytes(secret, "big")
            self.setStr(f"{int_key}".encode("utf-8"), 10)

    def get_str(self) -> str:
        return self.getStr(16).decode("utf-8")  # .zfill(64)


def new_private_key(sk: bytes = b"") -> PrivateKey:
    return PrivateKey(sk)


class KeyPair:
    def __init__(self, priv_key: PrivateKey = None) -> None:
        if not priv_key:
            self.priv_key = PrivateKey()
        else:
            self.priv_key = priv_key

        self.pub_g1 = bn256Utils.mul_by_generator_g1(self.priv_key).normalize()
        self.pub_g2 = bn256Utils.mul_by_generator_g2(self.priv_key).normalize()

    @staticmethod
    def from_string(sk: str, base=16) -> "KeyPair":
        pk = PrivateKey()
        pk.setStr(sk.encode("utf-8"), base)
        return KeyPair(pk)

    def save_to_file(self, _path: str, password: str):
        priv_key = "0x" + self.priv_key.getStr(16).decode("utf-8").rjust(64, "0")
        keystore_json = Account.encrypt(priv_key, password)
        keystore_json["pubKey"] = self.pub_g1.getStr().decode("utf-8")
        os.makedirs(os.path.dirname(_path), exist_ok=True)
        with open(_path, "w") as f:
            f.write(json.dumps(keystore_json))

    @staticmethod
    def read_from_file(_path: str, password: str):
        with open(_path, "r") as f:
            keystore_json = json.load(f)
        if "version" not in keystore_json:
            keystore_json["version"] = 3

        private_key = Account.decrypt(keystore_json, password)
        return KeyPair(PrivateKey(bytes(private_key)))

    def sign_message(self, msg_bytes: bytes) -> Signature:
        h = bn256Utils.map_to_curve(msg_bytes)
        return self.sign_hashed_to_curve_message(h)

    def sign_hashed_to_curve_message(self, msg_map_point: G1Point) -> Signature:
        sig = (msg_map_point * self.priv_key).normalize()
        return Signature.from_g1_point(sig)

    def get_pub_g1(self) -> G1Point:
        return bn256Utils.mul_by_generator_g1(self.priv_key)

    def get_pub_g2(self) -> G2Point:
        return bn256Utils.mul_by_generator_g2(self.priv_key)


def new_key_pair(priv_key: PrivateKey) -> KeyPair:
    return KeyPair(priv_key)


def new_key_pair_from_string(sk: str) -> KeyPair:
    return KeyPair.from_string(sk)


def gen_random_bls_keys() -> KeyPair:
    return KeyPair()


def g1_to_tupple(g1):
    return (int(g1.getX().getStr()), int(g1.getY().getStr()))


def g2_to_tupple(g2):
    return (
        (
            int(g2.getX().get_a().getStr()),
            int(g2.getX().get_b().getStr()),
        ),
        (
            int(g2.getY().get_a().getStr()),
            int(g2.getY().get_b().getStr()),
        ),
    )



import logging

from eth_typing import Address
from eth_utils import event_abi_to_log_topic
from web3 import Web3
from web3._utils.events import get_event_data
from web3.contract.contract import Contract

from eigensdk.chainio import utils
from eigensdk.chainio.utils import bitmap_to_quorum_ids
from eigensdk.crypto.bls.attestation import G1Point, G2Point
from eigensdk.types_ import (
    OperatorPubkeys,
    OperatorStateRetrieverCheckSignaturesIndices,
    OperatorStateRetrieverOperator,
    StakeRegistryTypesStrategyParams,
    StakeRegistryTypesStakeUpdate,
    BLSApkRegistryTypesApkUpdate,
)

DEFAULT_QUERY_BLOCK_RANGE = 10_000


class AvsRegistryReader:
    def __init__(
        self,
        registry_coordinator: Contract,
        registry_coordinator_addr: Address,
        bls_apk_registry: Contract,
        bls_apk_registry_addr: Address,
        operator_state_retriever: Contract,
        service_manager: Contract,
        stake_registry: Contract,
        logger: logging.Logger,
        eth_http_client: Web3,
    ):

        self.logger: logging.Logger = logger
        self.bls_apk_registry: Contract = bls_apk_registry
        self.bls_apk_registry_addr = bls_apk_registry_addr
        self.registry_coordinator: Contract = registry_coordinator
        self.registry_coordinator_addr = registry_coordinator_addr
        self.operator_state_retriever: Contract = operator_state_retriever
        self.service_manager = service_manager
        self.stake_registry: Contract = stake_registry
        self.eth_http_client: Web3 = eth_http_client

    def get_quorum_count(self) -> int:
        return self.registry_coordinator.functions.quorumCount().call()

    def get_operators_stake_in_quorums_at_current_block(
        self, quorum_numbers: list[int]
    ) -> list[list[OperatorStateRetrieverOperator]]:
        cur_block = self.eth_http_client.eth.block_number
        return self.get_operators_stake_in_quorums_at_block(quorum_numbers, cur_block)

    def get_operators_stake_in_quorums_at_block(
        self, quorum_numbers: list[int], block_number: int
    ) -> list[list[OperatorStateRetrieverOperator]]:
        operator_stakes = self.operator_state_retriever.get_function_by_signature(
            "getOperatorState(address,bytes,uint32)"
        )(
            registryCoordinator=self.registry_coordinator_addr,
            quorumNumbers=utils.nums_to_bytes(quorum_numbers),
            blockNumber=block_number,
        ).call()
        return [
            [
                OperatorStateRetrieverOperator(
                    operator=operator[0],
                    operator_id=operator[1],
                    stake=operator[2],
                )
                for operator in quorum
            ]
            for quorum in operator_stakes
        ]

    def get_operator_addrs_in_quorums_at_current_block(
        self, quorum_numbers: list[int]
    ) -> list[list[str]]:
        cur_block = self.eth_http_client.eth.block_number
        stakes = self.get_operators_stake_in_quorums_at_block(quorum_numbers, cur_block)
        return [[op.operator for op in quorum] for quorum in stakes]

    def get_operators_stake_in_quorums_of_operator_at_block(
        self, operator_id: bytes, block_number: int
    ) -> tuple[list[int], list[list[OperatorStateRetrieverOperator]]]:
        quorum_bitmap, operator_stakes = self.operator_state_retriever.get_function_by_signature(
            "getOperatorState(address,bytes32,uint32)"
        )(self.registry_coordinator_addr, operator_id, block_number).call()

        quorums = bitmap_to_quorum_ids(quorum_bitmap)
        return quorums, [
            [
                OperatorStateRetrieverOperator(
                    operator=operator[0],
                    operator_id=operator[1],
                    stake=operator[2],
                )
                for operator in quorum
            ]
            for quorum in operator_stakes
        ]

    def get_operators_stake_in_quorums_of_operator_at_current_block(
        self, operator_id: bytes
    ) -> tuple[list[int], list[list[OperatorStateRetrieverOperator]]]:
        cur_block = self.eth_http_client.eth.block_number
        return self.get_operators_stake_in_quorums_of_operator_at_block(operator_id, cur_block)

    def get_operator_stake_in_quorums_of_operator_at_current_block(
        self, operator_id: bytes
    ) -> dict[int, int]:
        quorum_bitmap = self.registry_coordinator.functions.getCurrentQuorumBitmap(
            operator_id
        ).call()
        quorums = bitmap_to_quorum_ids(quorum_bitmap)
        return {
            quorum: self.stake_registry.functions.getCurrentStake(operator_id, quorum).call()
            for quorum in quorums
        }

    def weight_of_operator_for_quorum(self, quorum_number: int, operator_addr: str) -> int:
        return self.stake_registry.functions.weightOfOperatorForQuorum(
            quorum_number, operator_addr
        ).call()

    def strategy_params_length(self, quorum_number: int) -> int:
        return self.stake_registry.functions.strategyParamsLength(quorum_number).call()

    def strategy_params_by_index(
        self, quorum_number: int, index: int
    ) -> StakeRegistryTypesStrategyParams:
        result = self.stake_registry.functions.strategyParamsByIndex(quorum_number, index).call()
        return StakeRegistryTypesStrategyParams(strategy=result[0], multiplier=result[1])

    def get_stake_history_length(self, operator_id: bytes, quorum_number: int) -> int:
        return self.stake_registry.functions.getStakeHistoryLength(
            operator_id, quorum_number
        ).call()

    def get_stake_history(
        self, operator_id: bytes, quorum_number: int
    ) -> list[StakeRegistryTypesStakeUpdate]:
        history = self.stake_registry.functions.getStakeHistory(operator_id, quorum_number).call()
        return [
            StakeRegistryTypesStakeUpdate(
                update_block_number=update[0], next_update_block_number=update[1], stake=update[2]
            )
            for update in history
        ]

    def get_latest_stake_update(
        self, operator_id: bytes, quorum_number: int
    ) -> StakeRegistryTypesStakeUpdate:
        update = self.stake_registry.functions.getLatestStakeUpdate(
            operator_id, quorum_number
        ).call()
        return StakeRegistryTypesStakeUpdate(
            update_block_number=update[0], next_update_block_number=update[1], stake=update[2]
        )

    def get_stake_update_at_index(
        self, operator_id: bytes, quorum_number: int, index: int
    ) -> StakeRegistryTypesStakeUpdate:
        update = self.stake_registry.functions.getStakeUpdateAtIndex(
            quorum_number, operator_id, index
        ).call()
        return StakeRegistryTypesStakeUpdate(
            update_block_number=update[0], next_update_block_number=update[1], stake=update[2]
        )

    def get_stake_at_block_number(
        self,
        operator_id: bytes,
        quorum_number: int,
        block_number: int,
    ) -> int:
        return self.stake_registry.functions.getStakeAtBlockNumber(
            operator_id, quorum_number, block_number
        ).call()

    def get_stake_update_index_at_block_number(
        self,
        operator_id: bytes,
        quorum_number: int,
        block_number: int,
    ) -> int:
        return self.stake_registry.functions.getStakeUpdateIndexAtBlockNumber(
            operator_id, quorum_number, block_number
        ).call()

    def get_total_stake_history_length(self, quorum_number: int) -> int:
        return self.stake_registry.functions.getTotalStakeHistoryLength(quorum_number).call()

    def get_check_signatures_indices(
        self,
        reference_block_number: int,
        quorum_numbers: list[int],
        non_signer_operator_ids: list[bytes],
    ) -> OperatorStateRetrieverCheckSignaturesIndices:
        quorum_bytes = utils.nums_to_bytes(quorum_numbers)
        result = self.operator_state_retriever.functions.getCheckSignaturesIndices(
            self.registry_coordinator_addr,
            reference_block_number,
            quorum_bytes,
            non_signer_operator_ids,
        ).call()

        return OperatorStateRetrieverCheckSignaturesIndices(
            result[0], result[1], result[2], result[3]
        )

    def get_current_total_stake(self, quorum_number: int) -> int:
        return self.stake_registry.functions.getCurrentTotalStake(quorum_number).call()

    def get_total_stake_update_at_index(
        self, quorum_number: int, index: int
    ) -> StakeRegistryTypesStakeUpdate:
        update = self.stake_registry.functions.getTotalStakeUpdateAtIndex(
            quorum_number, index
        ).call()
        return StakeRegistryTypesStakeUpdate(
            update_block_number=update[0], next_update_block_number=update[1], stake=update[2]
        )

    def get_total_stake_at_block_number_from_index(
        self, quorum_number: int, block_number: int, index: int
    ) -> int:
        return self.stake_registry.functions.getTotalStakeAtBlockNumberFromIndex(
            quorum_number, block_number, index
        ).call()

    def get_total_stake_indices_at_block_number(
        self, quorum_numbers: list[int], block_number: int
    ) -> list[int]:
        quorum_bytes = utils.nums_to_bytes(quorum_numbers)
        return self.stake_registry.functions.getTotalStakeIndicesAtBlockNumber(
            block_number, quorum_bytes
        ).call()

    def get_minimum_stake_for_quorum(self, quorum_number: int) -> int:
        return self.stake_registry.functions.minimumStakeForQuorum(quorum_number).call()

    def get_strategy_params_at_index(
        self, quorum_number: int, index: int
    ) -> StakeRegistryTypesStrategyParams:
        return self.stake_registry.functions.strategyParams(quorum_number, index).call()

    def get_strategy_per_quorum_at_index(self, quorum_number: int, index: int) -> str:
        return self.stake_registry.functions.strategiesPerQuorum(quorum_number, index).call()

    # TODO: IMPLEMENT TEST BASED ON THE AVS SERVICE MANAGER
    def get_restakeable_strategies(self) -> list[str]:
        return self.service_manager.functions.getRestakeableStrategies().call()

    # TODO: IMPLEMENT TEST BASED ON THE AVS SERVICE MANAGER
    def get_operator_restaked_strategies(self, operator: str) -> list[str]:
        return self.service_manager.functions.getOperatorRestakedStrategies(operator).call()

    def get_stake_type_per_quorum(self, quorum_number: int) -> int:
        return self.stake_registry.functions.stakeTypePerQuorum(quorum_number).call()

    def get_slashable_stake_look_ahead_per_quorum(self, quorum_number: int) -> int:
        return self.stake_registry.functions.slashableStakeLookAheadPerQuorum(quorum_number).call()

    def get_operator_id(self, operator_address: Address) -> bytes:
        operator_id = self.registry_coordinator.functions.getOperatorId(operator_address).call()
        return operator_id

    def get_operator_from_id(self, operator_id: bytes) -> str:
        return self.registry_coordinator.functions.getOperatorFromId(operator_id).call()

    def query_registration_detail(self, operator_address: Address) -> list[bool]:
        operator_id = self.get_operator_id(operator_address=operator_address)
        value = self.registry_coordinator.functions.getCurrentQuorumBitmap(operator_id).call()
        return [(value & (1 << i)) != 0 for i in range(value.bit_length())]

    def is_operator_registered(self, operator_address: str) -> bool:
        return self.registry_coordinator.functions.getOperatorStatus(operator_address).call() == 1

    def is_operator_set_quorum(self, quorum_number: int) -> bool:
        return self.stake_registry.functions.isOperatorSetQuorum(quorum_number).call()

    def get_operator_id_from_operator_address(self, operator_address: str) -> bytes:
        return self.bls_apk_registry.functions.operatorToPubkeyHash(operator_address).call()

    def get_operator_address_from_operator_id(self, operator_pubkey_hash: bytes) -> str:
        return self.bls_apk_registry.functions.pubkeyHashToOperator(operator_pubkey_hash).call()

    def get_pubkey_from_operator_address(self, operator_address: str) -> G1Point:
        operator_pubkey = self.bls_apk_registry.functions.operatorToPubkey(operator_address).call()
        return G1Point(operator_pubkey[0], operator_pubkey[1])

    def get_apk_update(self, quorum_number: int, index: int) -> BLSApkRegistryTypesApkUpdate:
        update = self.bls_apk_registry.functions.apkHistory(quorum_number, index).call()
        return BLSApkRegistryTypesApkUpdate(
            apk_hash=bytes(update[0]),
            update_block_number=update[1],
            next_update_block_number=update[2],
        )

    def get_current_apk(self, quorum_number: int) -> G1Point:
        apk = self.bls_apk_registry.functions.currentApk(quorum_number).call()
        return G1Point(x=apk[0], y=apk[1])

    def query_existing_registered_operator_sockets(
        self,
        start_block: int = 0,
        stop_block: int | None = None,
        block_range: int = DEFAULT_QUERY_BLOCK_RANGE,
    ) -> tuple[dict[bytes, str], int]:
        if stop_block is None:
            stop_block = self.eth_http_client.eth.block_number

        operator_id_to_socket_map: dict[bytes, str] = {}

        event_abi = self.registry_coordinator.events.OperatorSocketUpdate._get_event_abi()
        event_topic = event_abi_to_log_topic(event_abi)

        for i in range(start_block, stop_block + 1, block_range):
            to_block = min(i + block_range - 1, stop_block)

            try:
                logs = self.eth_http_client.eth.get_logs(
                    {
                        "fromBlock": i,
                        "toBlock": to_block,
                        "address": self.registry_coordinator.address,
                        "topics": [Web3.to_hex(event_topic)],
                    }
                )
            except Exception as e:
                self.logger.warning(f"Failed to fetch logs for blocks {i}-{to_block}: {e}")
                continue

            decoded_logs = [
                get_event_data(self.eth_http_client.codec, event_abi, log) for log in logs
            ]

            for log in decoded_logs:
                operator_id = log["args"]["operatorId"]
                socket = log["args"]["socket"]

                operator_id_to_socket_map[operator_id] = socket

            self.logger.debug(
                "avsRegistryChainReader.query_existing_registered_operator_sockets",
                extra={
                    "numTransactionLogs": len(decoded_logs),
                    "fromBlock": i,
                    "toBlock": to_block,
                },
            )

        return operator_id_to_socket_map, stop_block

    def query_existing_registered_operator_pubkeys(
        self,
        start_block: int = 0,
        stop_block: int | None = None,
        block_range: int = DEFAULT_QUERY_BLOCK_RANGE,
    ) -> tuple[list[Address], list[OperatorPubkeys]]:
        if stop_block is None:
            stop_block = self.eth_http_client.eth.block_number

        operator_pubkeys: list[OperatorPubkeys] = []
        operator_addresses: list[Address] = []
        for i in range(start_block, stop_block + 1, block_range):
            to_block: int = min(i + block_range - 1, stop_block)

            event_abi = self.bls_apk_registry.events.NewPubkeyRegistration._get_event_abi()
            event_topic = event_abi_to_log_topic(event_abi)
            logs = self.eth_http_client.eth.get_logs(
                {
                    "fromBlock": i,
                    "toBlock": to_block,
                    "address": self.bls_apk_registry.address,
                    "topics": [Web3.to_hex(event_topic)],
                }
            )

            pubkey_updates = [
                get_event_data(self.eth_http_client.codec, event_abi, log) for log in logs
            ]

            self.logger.debug(
                "avsRegistryChainReader.query_existing_registered_operator_pubkeys",
                extra={
                    "numTransactionLogs": len(pubkey_updates),
                    "from_block": i,
                    "to_block": to_block,
                },
            )
            for update in pubkey_updates:
                operator_addr = update["args"]["operator"]
                pubkey_g1 = update["args"]["pubkeyG1"]
                pubkey_g2 = update["args"]["pubkeyG2"]
                operator_pubkeys.append(
                    OperatorPubkeys(
                        g1_pub_key=G1Point(pubkey_g1["X"], pubkey_g1["Y"]),
                        g2_pub_key=G2Point(*pubkey_g2["X"], *pubkey_g2["Y"]),
                    )
                )
                operator_addresses.append(operator_addr)
        return operator_addresses, operator_pubkeys

    def get_registry_coordinator_owner(self) -> str:
        return self.registry_coordinator.functions.owner().call()

    def is_registry_coordinator_owner(self, address: str) -> bool:
        owner = self.get_registry_coordinator_owner()
        return owner.lower() == address.lower()

    def can_satisfy_only_coordinator_owner_modifier(self, address: str) -> bool:
        return self.is_registry_coordinator_owner(address)



import logging
from typing import List, Dict, Tuple

from eth_account.signers.local import LocalAccount
from eth_typing import Address
from web3 import Web3
from web3.contract.contract import Contract, ContractFunction
from web3.types import TxReceipt

from eigensdk.chainio.utils import Transactor, nums_to_bytes
from eigensdk.chainio.clients.elcontracts.reader import ELReader


class AvsRegistryWriter:
    def __init__(
        self,
        registry_coordinator: Contract,
        operator_state_retriever: Contract,
        service_manager: Contract,
        service_manager_addr: Address,
        stake_registry: Contract,
        bls_apk_registry: Contract,
        el_reader: ELReader,
        logger: logging.Logger,
        eth_http_client: Web3,
        pk_wallet: LocalAccount,
    ):

        self.registry_coordinator: Contract = registry_coordinator
        self.operator_state_retriever: Contract = operator_state_retriever
        self.service_manager: Contract = service_manager
        self.service_manager_addr: Address = service_manager_addr
        self.stake_registry: Contract = stake_registry
        self.bls_apk_registry: Contract = bls_apk_registry
        self.el_reader: ELReader = el_reader
        self.logger: logging.Logger = logger
        self.web3: Web3 = eth_http_client  # Create alias for compatibility
        self.transactor = Transactor(pk_wallet, eth_http_client)

    def send_transaction(self, func: ContractFunction):
        return self.transactor.send(func)

    def update_stakes_of_entire_operator_set_for_quorums(
        self,
        operators_per_quorum: List[List[str]],
        quorum_numbers: List[int],
    ) -> TxReceipt:
        func = self.registry_coordinator.functions.updateOperatorsForQuorum(
            operators_per_quorum,
            nums_to_bytes(quorum_numbers),
        )
        return self.send_transaction(func)

    def update_stakes_of_operator_subset_for_all_quorums(self, operators: List[str]) -> TxReceipt:
        func = self.registry_coordinator.functions.updateOperators(operators)
        return self.send_transaction(func)

    def update_socket(
        self,
        socket: str,
    ) -> TxReceipt:
        func = self.registry_coordinator.functions.updateSocket(socket)
        return self.send_transaction(func)

    def set_rewards_initiator(
        self,
        rewards_initiator_addr: str,
    ) -> TxReceipt:
        func = self.service_manager.functions.setRewardsInitiator(rewards_initiator_addr)
        return self.send_transaction(func)

    def set_slashable_stake_lookahead(
        self,
        quorum_number: int,
        look_ahead_period: int,
    ) -> TxReceipt:
        func = self.stake_registry.functions.setSlashableStakeLookahead(
            quorum_number, look_ahead_period
        )
        return self.send_transaction(func)

    def set_minimum_stake_for_quorum(
        self,
        quorum_number: int,
        minimum_stake: int,
    ) -> TxReceipt:
        func = self.stake_registry.functions.setMinimumStakeForQuorum(quorum_number, minimum_stake)
        return self.send_transaction(func)

    def create_total_delegated_stake_quorum(
        self,
        operator_set_params: Tuple[int, int, int],
        minimum_stake_required: int,
        strategy_params: List[Tuple[str, int]],
    ) -> TxReceipt:
        func = self.registry_coordinator.functions.createTotalDelegatedStakeQuorum(
            operator_set_params, minimum_stake_required, strategy_params
        )
        return self.send_transaction(func)

    def create_slashable_stake_quorum(
        self,
        operator_set_params: Tuple[int, int, int],
        minimum_stake_required: int,
        strategy_params: List[Tuple[str, int]],
        look_ahead_period: int,
    ) -> TxReceipt:
        func = self.registry_coordinator.functions.createSlashableStakeQuorum(
            operator_set_params,
            minimum_stake_required,
            strategy_params,
            look_ahead_period,
        )
        return self.send_transaction(func)

    def eject_operator(
        self,
        operator_address: str,
        quorum_numbers: List[int],
    ) -> TxReceipt:
        func = self.registry_coordinator.functions.ejectOperator(
            operator_address,
            nums_to_bytes(quorum_numbers),
        )
        return self.send_transaction(func)

    def set_operator_set_params(
        self,
        quorum_number: int,
        operator_set_params: Dict,
    ) -> TxReceipt:
        func = self.registry_coordinator.functions.setOperatorSetParams(
            quorum_number,
            operator_set_params,
        )
        return self.send_transaction(func)

    def set_churn_approver(
        self,
        churn_approver_address: str,
    ) -> TxReceipt:
        func = self.registry_coordinator.functions.setChurnApprover(churn_approver_address)
        return self.send_transaction(func)

    def set_ejector(
        self,
        ejector_address: str,
    ) -> TxReceipt:
        func = self.registry_coordinator.functions.setEjector(ejector_address)
        return self.send_transaction(func)

    def modify_strategy_params(
        self,
        quorum_number: int,
        strategy_indices: List[int],
        multipliers: List[int],
    ) -> TxReceipt:
        func = self.stake_registry.functions.modifyStrategyParams(
            quorum_number,
            strategy_indices,
            multipliers,
        )
        return self.send_transaction(func)

    def set_avs(self, avs_address: str) -> TxReceipt:
        func = self.registry_coordinator.functions.setAVS(avs_address)
        return self.send_transaction(func)

    def set_ejection_cooldown(
        self,
        ejection_cooldown: int,
    ) -> TxReceipt:
        func = self.registry_coordinator.functions.setEjectionCooldown(ejection_cooldown)
        return self.send_transaction(func)

    def add_strategies(
        self,
        quorum_number: int,
        strategy_params: List[Dict],
    ) -> TxReceipt:
        func = self.stake_registry.functions.addStrategies(quorum_number, strategy_params)
        return self.send_transaction(func)

    def update_avs_metadata_uri(
        self,
        metadata_uri: str,
    ) -> TxReceipt:
        func = self.service_manager.functions.updateAVSMetadataURI(metadata_uri)
        return self.send_transaction(func)

    def remove_strategies(
        self,
        quorum_number: int,
        indices_to_remove: List[int],
    ) -> TxReceipt:
        func = self.stake_registry.functions.removeStrategies(quorum_number, indices_to_remove)
        return self.send_transaction(func)

    def create_avs_rewards_submission(
        self,
        rewards_submission: List[Dict],
    ) -> TxReceipt:
        func = self.service_manager.functions.createAVSRewardsSubmission(rewards_submission)
        return self.send_transaction(func)

    def create_operator_directed_avs_rewards_submission(
        self,
        operator_directed_rewards_submissions: List[Dict],
    ) -> TxReceipt:
        func = self.service_manager.functions.createOperatorDirectedAVSRewardsSubmission(
            operator_directed_rewards_submissions
        )
        return self.send_transaction(func)



import logging
from typing import Tuple, List, Any, Dict, Optional, Union, Sequence

from eth_typing import Address
from web3 import Web3
from web3.contract.contract import Contract
from web3.types import ChecksumAddress


class ELReader:
    def __init__(
        self,
        allocation_manager: Contract,
        avs_directory: Contract,
        delegation_manager: Contract,
        permission_controller: Contract,
        reward_coordinator: Contract,
        strategy_manager: Contract,
        logger: logging.Logger,
        eth_http_client: Web3,
        strategy_abi: List[Dict[str, Any]],
        erc20_abi: List[Dict[str, Any]],
    ):
        self.allocation_manager = allocation_manager
        self.avs_directory = avs_directory
        self.delegation_manager = delegation_manager
        self.permission_controller = permission_controller
        self.reward_coordinator = reward_coordinator
        self.strategy_manager = strategy_manager
        self.eth_http_client = eth_http_client
        self.logger = logger
        self.strategy_abi = strategy_abi
        self.erc20_abi = erc20_abi

        if allocation_manager is None:
            self.logger.warning("AllocationManager contract not provided")

        if avs_directory is None:
            self.logger.warning("AvsDirectory contract not provided")

        if delegation_manager is None:
            self.logger.warning("DelegationManager contract not provided")

        if permission_controller is None:
            self.logger.warning("PermissionController contract not provided")

        if reward_coordinator is None:
            self.logger.warning("RewardCoordinator contract not provided")

        if strategy_manager is None:
            self.logger.warning("StrategyManager contract not provided")

        if strategy_abi is None:
            self.logger.warning("Strategy ABI not provided")

        if erc20_abi is None:
            self.logger.warning("ERC20 ABI not provided")

    def get_allocatable_magnitude(
        self,
        operator_addr: Optional[Union[Address, ChecksumAddress, str]],
        strategy_addr: Optional[Union[Address, ChecksumAddress, str]],
    ) -> int:

        return self.allocation_manager.functions.getAllocatableMagnitude(
            operator_addr, strategy_addr
        ).call()

    def get_max_magnitudes(
        self,
        operator_addr: Optional[Union[Address, ChecksumAddress, str]],
        strategy_addrs: Sequence[Optional[Union[Address, ChecksumAddress, str]]],
    ) -> List[int]:

        return self.allocation_manager.functions.getMaxMagnitudes(
            operator_addr, strategy_addrs
        ).call()

    def get_allocation_info(
        self,
        operator_addr: Optional[Union[Address, ChecksumAddress, str]],
        strategy_addr: Optional[Union[Address, ChecksumAddress, str]],
    ) -> List[Dict[str, Any]]:
        return [
            {
                "OperatorSetId": sid,
                "AvsAddress": avs,
                "CurrentMagnitude": mag,
                "PendingDiff": diff,
                "EffectBlock": block,
            }
            for (sid, avs), (mag, diff, block) in zip(
                *self.allocation_manager.functions.getStrategyAllocations(
                    operator_addr, strategy_addr
                ).call()
            )
        ]

    def get_operator_shares(
        self,
        operator_address: Optional[Union[Address, ChecksumAddress, str]],
        strategy_addresses: Sequence[Union[Address, ChecksumAddress]],
    ) -> List[int]:
        return self.delegation_manager.functions.getOperatorShares(
            operator_address, strategy_addresses
        ).call()

    def get_operator_sets_for_operator(
        self, operator_addr: Optional[Union[Address, str]]
    ) -> List[Dict[str, Any]]:
        return [
            {"Id": s[0], "AvsAddress": s[1]}
            for s in self.allocation_manager.functions.getAllocatedSets(operator_addr).call()
        ]

    def get_allocation_delay(self, operator_addr: Optional[Union[Address, str]]) -> int:
        is_set, delay = self.allocation_manager.functions.getAllocationDelay(operator_addr).call()
        return delay if is_set else 0

    def get_registered_sets(
        self, operator_addr: Optional[Union[Address, str]]
    ) -> List[Dict[str, Any]]:
        return [
            {"Id": s[0], "Avs": s[1]}
            for s in self.allocation_manager.functions.getRegisteredSets(operator_addr).call()
        ]

    def is_operator_registered_with_avs(
        self,
        operator_address: Optional[Union[Address, str]],
        avs_address: Optional[Union[Address, str]],
    ) -> bool:
        return (
            self.avs_directory.functions.avsOperatorStatus(avs_address, operator_address).call()
            == 1
        )

    def is_operator_registered_with_operator_set(
        self, operator_addr: Optional[Union[Address, str]], operator_set: dict
    ) -> bool:
        return any(
            s[0] == operator_set.get("Id", 0) and s[1] == operator_set.get("Avs")
            for s in self.allocation_manager.functions.getRegisteredSets(operator_addr).call()
        )

    def is_operator_slashable(
        self, operator_address: Optional[Union[Address, str]], operator_set: Dict[str, Any]
    ) -> bool:
        return self.allocation_manager.functions.isOperatorSlashable(
            operator_address,
            (Web3.to_checksum_address(operator_set["Avs"]), operator_set["Id"]),
        ).call()

    def get_allocated_stake(
        self,
        operator_set: Dict[str, Any],
        operator_addresses: Sequence[Union[Address, ChecksumAddress]],
        strategy_addresses: Sequence[Union[Address, ChecksumAddress]],
    ) -> List[List[int]]:
        stakes = self.allocation_manager.functions.getAllocatedStake(
            (Web3.to_checksum_address(operator_set["Avs"]), operator_set["Id"]),
            operator_addresses,
            strategy_addresses,
        ).call()
        return stakes

    def get_operators_for_operator_set(self, operator_set: dict) -> list:

        if operator_set.get("id", 0) == 0:
            raise ValueError("Legacy AVSs not supported")

        if not self.allocation_manager:
            raise ValueError("AllocationManager contract not provided")

        operator_set_tuple = (
            Web3.to_checksum_address(operator_set["avs"]),
            operator_set.get("quorumNumber", 0),
        )

        operators = self.allocation_manager.functions.getMembers(operator_set_tuple).call()

        return operators

    def get_num_operators_for_operator_set(self, operator_set: dict) -> int:
        return self.allocation_manager.functions.getMemberCount(
            (Web3.to_checksum_address(operator_set["Avs"]), operator_set["Id"])
        ).call()

    def get_strategies_for_operator_set(self, operator_set: dict) -> List[Address]:

        if operator_set.get("id", 0) == 0:
            raise ValueError("Legacy AVSs not supported")

        if not self.allocation_manager:
            raise ValueError("AllocationManager contract not provided")

        # Create the operator set tuple expected by the contract
        operator_set_tuple = (
            Web3.to_checksum_address(operator_set["avs"]),
            operator_set.get("quorumNumber", 0),
        )

        # Call the contract function
        strategies = self.allocation_manager.functions.getStrategiesInOperatorSet(
            operator_set_tuple
        ).call()

        return strategies

    def is_operator_registered(self, operator_address: Optional[Union[Address, str]]) -> bool:
        return self.delegation_manager.functions.isOperator(operator_address).call()

    def get_staker_shares(
        self, staker_address: Optional[Union[Address, str]]
    ) -> Tuple[List[Address], List[int]]:
        return self.delegation_manager.functions.getDepositedShares(staker_address).call()

    def get_avs_registrar(self, avs_address: Optional[Union[Address, str]]) -> Address:
        return self.allocation_manager.functions.getAVSRegistrar(avs_address).call()

    def get_delegated_operator(
        self, staker_address: Optional[Union[Address, str]], block_number: Optional[int] = None
    ) -> Address:
        return (
            self.delegation_manager.functions.delegatedTo(staker_address).call(
                block_identifier=block_number
            )
            if block_number
            else self.delegation_manager.functions.delegatedTo(staker_address).call()
        )

    def get_operator_details(self, operator: Dict[str, Any]) -> Dict[str, Any]:
        addr = Web3.to_checksum_address(operator["Address"])
        is_set, delay = self.allocation_manager.functions.getAllocationDelay(addr).call()
        return dict(
            Address=addr,
            DelegationApproverAddress=(
                self.delegation_manager.functions.delegationApproverSaltIsSpent(
                    addr, b"\x00" * 32
                ).call()
            ),
            AllocationDelay=delay if is_set else 0,
        )

    def get_operator_shares_in_strategy(
        self,
        operator_addr: Optional[Union[Address, str]],
        strategy_addr: Optional[Union[Address, str]],
    ) -> int:
        return self.delegation_manager.functions.operatorShares(operator_addr, strategy_addr).call()

    def calculate_delegation_approval_digest_hash(
        self,
        staker: Optional[Union[Address, str]],
        operator: Optional[Union[Address, str]],
        delegation_approver: Optional[Union[Address, str]],
        approver_salt: bytes,
        expiry: int,
    ) -> bytes:
        return self.delegation_manager.functions.calculateDelegationApprovalDigestHash(
            staker, operator, delegation_approver, approver_salt, expiry
        ).call()

    def get_operators_shares(
        self,
        operator_addresses: Sequence[Union[Address, ChecksumAddress]],
        strategy_addresses: Sequence[Union[Address, ChecksumAddress]],
    ) -> List[List[int]]:
        return self.delegation_manager.functions.getOperatorsShares(
            operator_addresses, strategy_addresses
        ).call()

    def get_delegation_approver_salt_is_spent(
        self, delegation_approver: Optional[Union[Address, str]], approver_salt: bytes
    ) -> bool:
        return self.delegation_manager.functions.delegationApproverSaltIsSpent(
            delegation_approver, approver_salt
        ).call()

    def get_pending_withdrawal_status(self, withdrawal_root: bytes) -> bool:
        return self.delegation_manager.functions.pendingWithdrawals(withdrawal_root).call()

    def get_cumulative_withdrawals_queued(self, staker: Optional[Union[Address, str]]) -> int:
        return self.delegation_manager.functions.cumulativeWithdrawalsQueued(staker).call()

    def can_call(
        self,
        account_address: Optional[Union[Address, str]],
        appointee_address: Optional[Union[Address, str]],
        target: Optional[Union[Address, str]],
        selector: bytes,
    ) -> bool:
        return self.permission_controller.functions.canCall(
            account_address, appointee_address, target, selector
        ).call()

    def list_appointees(
        self,
        account_address: Optional[Union[Address, str]],
        target: Optional[Union[Address, str]],
        selector: bytes,
    ) -> List[Address]:
        return self.permission_controller.functions.getAppointees(
            account_address, target, selector
        ).call()

    def list_appointee_permissions(
        self,
        account_address: Optional[Union[Address, str]],
        appointee_address: Optional[Union[Address, str]],
    ) -> Tuple[List[Address], List[bytes]]:
        return self.permission_controller.functions.getAppointeePermissions(
            account_address, appointee_address
        ).call()

    def list_pending_admins(self, account_address: Optional[Union[Address, str]]) -> List[Address]:
        return self.permission_controller.functions.getPendingAdmins(account_address).call()

    def list_admins(self, account_address: Optional[Union[Address, str]]) -> List[Address]:
        return self.permission_controller.functions.getAdmins(account_address).call()

    def is_pending_admin(
        self,
        account_address: Optional[Union[Address, str]],
        pending_admin_address: Optional[Union[Address, str]],
    ) -> bool:
        return self.permission_controller.functions.isPendingAdmin(
            account_address, pending_admin_address
        ).call()

    def is_admin(
        self,
        account_address: Optional[Union[Address, str]],
        admin_address: Optional[Union[Address, str]],
    ) -> bool:
        return self.permission_controller.functions.isAdmin(account_address, admin_address).call()

    def get_distribution_roots_length(self) -> int:
        return self.reward_coordinator.functions.getDistributionRootsLength().call()

    def curr_rewards_calculation_end_timestamp(self) -> int:
        return self.reward_coordinator.functions.currRewardsCalculationEndTimestamp().call()

    def get_current_claimable_distribution_root(self) -> Dict[str, Any]:
        return dict(
            zip(
                ["root", "startBlock", "endBlock", "totalClaimable"],
                self.reward_coordinator.functions.getCurrentClaimableDistributionRoot().call(),
            )
        )

    def get_root_index_from_hash(self, root_hash: bytes) -> int:
        return self.reward_coordinator.functions.getRootIndexFromHash(root_hash).call()

    def get_cumulative_claimed(
        self, earner: Optional[Union[Address, str]], token: Optional[Union[Address, str]]
    ) -> int:
        return self.reward_coordinator.functions.cumulativeClaimed(earner, token).call()

    def check_claim(self, claim: Dict[str, Any]) -> bool:
        token_indices = claim.get("tokenIndices", [])
        token_tree_proofs = claim.get("tokenTreeProofs", [])
        token_leaves = claim.get("tokenLeaves", [])
        if not (len(token_indices) == len(token_tree_proofs) == len(token_leaves)):
            raise ValueError(
                "tokenIndices, tokenTreeProofs, and tokenLeaves must have the same length"
            )
        root_index = claim.get("rootIndex", 0)
        distribution_roots_length = self.get_distribution_roots_length()
        if distribution_roots_length == 0:
            raise ValueError("No distribution roots exist in the contract yet")
        if root_index < 0 or root_index >= distribution_roots_length:
            raise ValueError(
                f"""rootIndex {root_index} is out of bounds. Must be
                between 0 and {distribution_roots_length - 1}"""
            )
        earner_leaf = claim.get("earnerLeaf", {})
        if not earner_leaf.get("earner") or not Web3.is_address(earner_leaf.get("earner")):
            raise ValueError("Invalid earner address in earnerLeaf")
        if not earner_leaf.get("earnerTokenRoot") or len(earner_leaf.get("earnerTokenRoot")) != 32:
            raise ValueError("earnerTokenRoot must be 32 bytes")
        earner_leaf_tuple = (
            Web3.to_checksum_address(earner_leaf.get("earner")),
            earner_leaf.get("earnerTokenRoot"),
        )
        token_leaves_tuples = []
        for i, leaf in enumerate(token_leaves):
            if not leaf.get("token") or not Web3.is_address(leaf.get("token")):
                raise ValueError(f"Invalid token address in tokenLeaves[{i}]")
            if not isinstance(leaf.get("cumulativeEarnings"), int):
                raise ValueError(f"cumulativeEarnings must be an integer in tokenLeaves[{i}]")

            token_leaves_tuples.append(
                (
                    Web3.to_checksum_address(leaf.get("token")),
                    leaf.get("cumulativeEarnings", 0),
                )
            )
        claim_tuple = (
            root_index,
            claim.get("earnerIndex", 0),
            claim.get("earnerTreeProof", b""),
            earner_leaf_tuple,
            token_indices,
            token_tree_proofs,
            token_leaves_tuples,
        )
        return self.reward_coordinator.functions.checkClaim(claim_tuple).call()

    def get_operator_avs_split(
        self, operator: Optional[Union[Address, str]], avs: Optional[Union[Address, str]]
    ) -> int:
        return self.reward_coordinator.functions.getOperatorAVSSplit(operator, avs).call()

    def get_operator_pi_split(self, operator: Optional[Union[Address, str]]) -> int:
        return self.reward_coordinator.functions.getOperatorPISplit(operator).call()

    def get_operator_set_split(
        self, operator: Optional[Union[Address, str]], operator_set: dict
    ) -> int:
        return self.reward_coordinator.functions.getOperatorSetSplit(
            operator, (operator_set["Avs"], operator_set["Id"])
        ).call()

    def get_curr_rewards_calculation_end_timestamp(self) -> int:
        return self.reward_coordinator.functions.currRewardsCalculationEndTimestamp().call()

    def get_rewards_updater(self) -> Optional[Union[Address, str]]:
        return self.reward_coordinator.functions.rewardsUpdater().call()

    def get_default_operator_split_bips(self) -> int:
        return self.reward_coordinator.functions.defaultOperatorSplitBips().call()

    def get_claimer_for(self, earner: Optional[Union[Address, str]]) -> Address:
        return self.reward_coordinator.functions.claimerFor(earner).call()

    def get_submission_nonce(self, avs: Optional[Union[Address, str]]) -> int:
        return self.reward_coordinator.functions.submissionNonce(avs).call()

    def get_is_avs_rewards_submission_hash(
        self, avs: Optional[Union[Address, str]], hash: bytes
    ) -> bool:
        return self.reward_coordinator.functions.isAVSRewardsSubmissionHash(avs, hash).call()

    def get_is_rewards_submission_for_all_hash(
        self, avs: Optional[Union[Address, str]], hash: bytes
    ) -> bool:
        return self.reward_coordinator.functions.isRewardsSubmissionForAllHash(avs, hash).call()

    def get_is_rewards_for_all_submitter(self, submitter: Optional[Union[Address, str]]) -> bool:
        return self.reward_coordinator.functions.isRewardsForAllSubmitter(submitter).call()

    def get_is_rewards_submission_for_all_earners_hash(
        self, avs: Optional[Union[Address, str]], hash: bytes
    ) -> bool:
        return self.reward_coordinator.functions.isRewardsSubmissionForAllEarnersHash(
            avs, hash
        ).call()

    def get_is_operator_directed_avs_rewards_submission_hash(
        self, avs: Optional[Union[Address, str]], hash: bytes
    ) -> bool:
        return self.reward_coordinator.functions.isOperatorDirectedAVSRewardsSubmissionHash(
            avs, hash
        ).call()

    def get_is_operator_directed_operator_set_rewards_submission_hash(
        self, avs: Optional[Union[Address, str]], hash: bytes
    ) -> bool:
        return self.reward_coordinator.functions.isOperatorDirectedOperatorSetRewardsSubmissionHash(
            avs, hash
        ).call()

    def get_strategy_and_underlying_token(
        self, strategy_addr: Union[Address, ChecksumAddress]
    ) -> Tuple[Contract, Optional[str]]:
        sc = self.eth_http_client.eth.contract(address=strategy_addr, abi=self.strategy_abi)
        return sc, sc.functions.underlyingToken().call()

    def get_strategy_and_underlying_erc20_token(
        self, strategy_addr: Union[Address, ChecksumAddress]
    ) -> Tuple[Contract, Contract, Optional[str]]:
        strategy_contract = self.eth_http_client.eth.contract(
            address=Web3.to_checksum_address(strategy_addr), abi=self.strategy_abi
        )

        token_addr = Web3.to_checksum_address(strategy_contract.functions.underlyingToken().call())
        return (
            strategy_contract,
            self.eth_http_client.eth.contract(token_addr, abi=self.erc20_abi),
            token_addr,
        )

    def calculate_operator_avs_registration_digest_hash(
        self,
        operator: Optional[Union[Address, str]],
        avs: Optional[Union[Address, str]],
        salt: bytes,
        expiry: int,
    ) -> bytes:
        return self.avs_directory.functions.calculateOperatorAVSRegistrationDigestHash(
            operator, avs, salt, expiry
        ).call()

    def get_encumbered_magnitude(
        self,
        operator_address: Optional[Union[Address, str]],
        strategy_address: Optional[Union[Address, str]],
    ) -> int:
        return self.allocation_manager.functions.getEncumberedMagnitude(
            operator_address, strategy_address
        ).call()

    def get_calculation_interval_seconds(self) -> int:
        return self.reward_coordinator.functions.CALCULATION_INTERVAL_SECONDS().call()

    def get_max_rewards_duration(self) -> int:
        return self.reward_coordinator.functions.MAX_REWARDS_DURATION().call()

    def get_max_retroactive_length(self) -> int:
        return self.reward_coordinator.functions.MAX_RETROACTIVE_LENGTH().call()

    def get_max_future_length(self) -> int:
        return self.reward_coordinator.functions.MAX_FUTURE_LENGTH().call()

    def get_genesis_rewards_timestamp(self) -> int:
        return self.reward_coordinator.functions.GENESIS_REWARDS_TIMESTAMP().call()

    def get_activation_delay(self) -> int:
        return self.reward_coordinator.functions.activationDelay().call()

    def get_deallocation_delay(self) -> int:
        return self.allocation_manager.functions.DEALLOCATION_DELAY().call()

    def get_allocation_configuration_delay(self) -> int:
        return self.allocation_manager.functions.ALLOCATION_CONFIGURATION_DELAY().call()

    def get_num_operator_sets_for_operator(
        self, operator_address: Optional[Union[Address, str]]
    ) -> int:
        return self.allocation_manager.functions.getAllocatedSets(operator_address).call()

    def get_slashable_shares(
        self,
        operator_address: Optional[Union[Address, ChecksumAddress, str]],
        operator_set: Dict[str, Any],
        strategies: Sequence[Union[Address, ChecksumAddress]],
    ) -> Optional[Dict[str, int]]:
        if operator_address is None:
            return None
        return self.allocation_manager.functions.getMinimumSlashableStake(
            operator_set, [operator_address], strategies, self.eth_http_client.eth.block_number
        ).call()

    def get_slashable_shares_for_operator_sets_before(
        self, operator_sets: List[Dict[str, Any]], future_block: int
    ) -> Optional[List[Dict[str, Any]]]:
        result = []
        for op_set in operator_sets:
            operators = self.get_operators_for_operator_set(op_set)
            strategies = self.get_strategies_for_operator_set(op_set)

            stakes = self.allocation_manager.functions.getMinimumSlashableStake(
                op_set,
                operators,
                strategies,
                future_block,
            ).call()

            result.append(
                {
                    "OperatorSet": op_set,
                    "Strategies": strategies,
                    "Operators": operators,
                    "SlashableStakes": stakes,
                }
            )
        return result

    def get_slashable_shares_for_operator_sets(
        self, operator_sets: List[Dict[str, Any]]
    ) -> Optional[List[Dict[str, Any]]]:
        return self.get_slashable_shares_for_operator_sets_before(
            operator_sets, self.eth_http_client.eth.block_number
        )



import json
import os


def load_abi(file_name):
    base_dir = os.path.dirname(__file__)
    abi_dir = os.path.join(base_dir, "ABI")
    abi_path = os.path.join(abi_dir, file_name)

    if not os.path.exists(abi_path):
        raise FileNotFoundError(f"ABI file not found: {abi_path}")

    with open(abi_path, "r") as file:
        data = json.load(file)
        return data["abi"]


ALLOCATION_MANAGER_ABI = load_abi("AllocationManager.json")
AVS_DIRECTORY_ABI = load_abi("AVSDirectory.json")
BLS_APK_REGISTRY_ABI = load_abi("BLSApkRegistry.json")
DELEGATION_MANAGER_ABI = load_abi("DelegationManager.json")
IERC20_ABI = load_abi("IERC20.json")
I_STRATEGY_ABI = load_abi("IStrategy.json")
OPERATOR_STATE_RETRIEVER_ABI = load_abi("OperatorStateRetriever.json")
PERMISSION_CONTROLLER_ABI = load_abi("PermissionController.json")
REGISTRY_COORDINATOR_ABI = load_abi("RegistryCoordinator.json")
REWARDS_COORDINATOR_ABI = load_abi("RewardsCoordinator.json")
SERVICE_MANAGER_BASE_ABI = load_abi("ServiceManagerBase.json")
STAKE_REGISTRY_ABI = load_abi("StakeRegistry.json")
STRATEGY_MANAGER_ABI = load_abi("StrategyManager.json")



from web3 import Web3

from eigensdk.chainio.utils import nums_to_bytes
from tests.builder import clients, config


def test_get_allocatable_magnitude():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    result = clients.el_reader.get_allocatable_magnitude(operator_addr, strategy_addr)
    assert isinstance(result, int)
    print(f"Allocatable magnitude: {result}")


def test_get_max_magnitudes():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategy_addresses = [Web3.to_checksum_address(config["strategy_addr"])]
    result = clients.el_reader.get_max_magnitudes(operator_addr, strategy_addresses)
    assert isinstance(result, list)
    assert all(isinstance(magnitude, int) for magnitude in result)
    print(f"Max magnitudes: {result}")


def test_get_allocation_info():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    result = clients.el_reader.get_allocation_info(operator_addr, strategy_addr)
    assert isinstance(result, list)
    for allocation in result:
        assert isinstance(allocation, dict)
        assert "OperatorSetId" in allocation
        assert "AvsAddress" in allocation
        assert "CurrentMagnitude" in allocation
        assert "PendingDiff" in allocation
        assert "EffectBlock" in allocation
    print(f"Allocation info: {result}")


def test_get_operator_shares():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategy_addresses = [Web3.to_checksum_address(config["strategy_addr"])]
    result = clients.el_reader.get_operator_shares(operator_addr, strategy_addresses)
    assert isinstance(result, list)
    assert all(isinstance(share, int) for share in result)
    print(f"Operator shares: {result}")


def test_get_operator_sets_for_operator():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.get_operator_sets_for_operator(operator_addr)
    assert isinstance(result, list)
    for operator_set in result:
        assert isinstance(operator_set, dict)
        assert "Id" in operator_set
        assert "AvsAddress" in operator_set
    print(f"Operator sets: {result}")


def test_get_allocation_delay():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.get_allocation_delay(operator_addr)
    assert isinstance(result, int)
    print(f"Allocation delay: {result}")


def test_get_registered_sets():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.get_registered_sets(operator_addr)
    assert isinstance(result, list)
    for registered_set in result:
        assert isinstance(registered_set, dict)
        assert "Id" in registered_set
        assert "Avs" in registered_set
    print(f"Registered sets: {result}")


def test_is_operator_registered_with_avs():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    result = clients.el_reader.is_operator_registered_with_avs(operator_addr, avs_addr)
    assert isinstance(result, bool)
    print(f"Is operator registered with AVS: {result}")


def test_is_operator_registered_with_operator_set():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    registered_sets = clients.el_reader.get_registered_sets(operator_addr)
    if registered_sets:
        operator_set = registered_sets[0]
        expected_result = True
    else:
        operator_set = {
            "Id": 0,
            "Avs": Web3.to_checksum_address(config["avs_registry_coordinator_address"]),
        }
        expected_result = False
    result = clients.el_reader.is_operator_registered_with_operator_set(operator_addr, operator_set)
    assert isinstance(result, bool)
    assert result == expected_result
    print(f"Is operator registered with operator set: {result} (Expected: {expected_result})")


def test_is_operator_slashable():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    operator_sets = clients.el_reader.get_operator_sets_for_operator(operator_addr)
    if operator_sets:
        operator_set = operator_sets[0]
    else:
        operator_set = {
            "Id": 0,
            "Avs": Web3.to_checksum_address(config["avs_registry_coordinator_address"]),
        }
    result = clients.el_reader.is_operator_slashable(operator_addr, operator_set)
    assert isinstance(result, bool)
    print(f"Is operator slashable: {result}")


def test_get_allocated_stake():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    operator_sets = clients.el_reader.get_operator_sets_for_operator(operator_addr)
    if operator_sets:
        operator_set = operator_sets[0]
    else:
        operator_set = {
            "Id": 0,
            "Avs": Web3.to_checksum_address(config["avs_registry_coordinator_address"]),
        }
    result = clients.el_reader.get_allocated_stake(operator_set, [operator_addr], [strategy_addr])
    assert isinstance(result, list)
    if result:
        assert all(isinstance(inner_list, list) for inner_list in result)
        for inner_list in result:
            assert all(isinstance(stake, int) for stake in inner_list)
    print(f"Allocated stake: {result}")


def test_get_operators_for_operator_set():
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    operator_set = {"id": 1, "avs": avs_addr, "quorumNumber": 1}
    result = clients.el_reader.get_operators_for_operator_set(operator_set)
    assert isinstance(result, list)
    for operator in result:
        assert Web3.is_address(operator)
    print(f"Operators for operator set: {result}")


def test_get_num_operators_for_operator_set():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    operator_sets = clients.el_reader.get_operator_sets_for_operator(operator_addr)
    if operator_sets:
        operator_set = operator_sets[0]
    else:
        operator_set = {
            "Id": 0,
            "Avs": Web3.to_checksum_address(config["avs_registry_coordinator_address"]),
        }
    result = clients.el_reader.get_num_operators_for_operator_set(operator_set)
    assert isinstance(result, int)
    print(f"Number of operators for operator set: {result}")


def test_get_strategies_for_operator_set():
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    operator_set = {"id": 1, "avs": avs_addr, "quorumNumber": 1}
    result = clients.el_reader.get_strategies_for_operator_set(operator_set)
    assert isinstance(result, list)
    for strategy in result:
        assert Web3.is_address(strategy)
    print(f"Strategies for operator set: {result}")


def test_is_operator_registered():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.is_operator_registered(operator_addr)
    assert isinstance(result, bool)
    print(f"Is operator registered: {result}")


def test_get_staker_shares():
    staker_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategies, shares = clients.el_reader.get_staker_shares(staker_addr)
    assert isinstance(strategies, list)
    assert isinstance(shares, list)
    assert len(strategies) == len(shares)
    for strategy in strategies:
        assert Web3.is_address(strategy)
    for share in shares:
        assert isinstance(share, int)
    print(f"Staker shares: strategies={strategies}, shares={shares}")


def test_get_avs_registrar():
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    result = clients.el_reader.get_avs_registrar(avs_addr)
    assert Web3.is_address(result)
    print(f"AVS registrar: {result.hex() if isinstance(result, bytes) else result}")


def test_get_delegated_operator():
    staker_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.get_delegated_operator(staker_addr)
    assert Web3.is_address(result) or result == "0x0000000000000000000000000000000000000000"
    print(f"Delegated operator: {result.hex() if isinstance(result, bytes) else result}")
    current_block = clients.eth_http_client.eth.block_number
    result_with_block = clients.el_reader.get_delegated_operator(staker_addr, current_block)
    assert (
        Web3.is_address(result_with_block)
        or result_with_block == "0x0000000000000000000000000000000000000000"
    )
    print(
        f"Delegated operator at block {current_block}: {result_with_block.hex() if isinstance(result_with_block, bytes) else result_with_block}"
    )


def test_get_operator_details():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    operator = {"Address": operator_addr}
    result = clients.el_reader.get_operator_details(operator)
    assert isinstance(result, dict)
    assert "Address" in result
    assert "DelegationApproverAddress" in result
    assert "AllocationDelay" in result
    assert Web3.is_address(result["Address"])
    assert isinstance(result["AllocationDelay"], int)
    print(f"Operator details: {result}")


def test_get_operator_shares_in_strategy():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    result = clients.el_reader.get_operator_shares_in_strategy(operator_addr, strategy_addr)
    assert isinstance(result, int)
    print(f"Operator shares in strategy: {result}")


def test_calculate_delegation_approval_digest_hash():
    staker_addr = Web3.to_checksum_address(config["operator_address_1"])
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    delegation_approver = Web3.to_checksum_address("0x0000000000000000000000000000000000000000")
    approver_salt = nums_to_bytes([0] * 32)
    expiry = 2**32 - 1
    result = clients.el_reader.calculate_delegation_approval_digest_hash(
        staker_addr, operator_addr, delegation_approver, approver_salt, expiry
    )
    assert isinstance(result, bytes)
    assert len(result) == 32
    print(f"Delegation approval digest hash: 0x{result.hex()}")


def test_get_operators_shares():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    result = clients.el_reader.get_operators_shares([operator_addr], [strategy_addr])
    assert isinstance(result, list)
    assert all(isinstance(inner_list, list) for inner_list in result)
    for inner_list in result:
        assert all(isinstance(share, int) for share in inner_list)
    print(f"Operators shares: {result}")


def test_get_delegation_approver_salt_is_spent():
    delegation_approver = Web3.to_checksum_address(config["operator_address_1"])
    approver_salt = nums_to_bytes([0] * 32)
    result = clients.el_reader.get_delegation_approver_salt_is_spent(
        delegation_approver, approver_salt
    )
    assert isinstance(result, bool)
    print(f"Delegation approver salt is spent: {result}")


def test_get_pending_withdrawal_status():
    withdrawal_root = nums_to_bytes([0] * 32)
    result = clients.el_reader.get_pending_withdrawal_status(withdrawal_root)
    assert isinstance(result, bool)
    print(f"Pending withdrawal status: {result}")


def test_get_cumulative_withdrawals_queued():
    staker_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.get_cumulative_withdrawals_queued(staker_addr)
    assert isinstance(result, int)
    print(f"Cumulative withdrawals queued: {result}")


def test_can_call():
    account_addr = Web3.to_checksum_address(config["operator_address_1"])
    appointee_addr = Web3.to_checksum_address(config["operator_address_1"])
    target_addr = Web3.to_checksum_address(config["avs_registry_coordinator_address"])
    selector = nums_to_bytes([12, 34, 56, 78])
    result = clients.el_reader.can_call(account_addr, appointee_addr, target_addr, selector)
    assert isinstance(result, bool)
    print(f"Can call: {result}")


def test_list_appointees():
    account_addr = Web3.to_checksum_address(config["operator_address_1"])
    target_addr = Web3.to_checksum_address(config["avs_registry_coordinator_address"])
    selector = nums_to_bytes([12, 34, 56, 78])
    result = clients.el_reader.list_appointees(account_addr, target_addr, selector)
    assert isinstance(result, list)
    for appointee in result:
        assert Web3.is_address(appointee)
    print(f"Appointees: {result}")


def test_list_appointee_permissions():
    account_addr = Web3.to_checksum_address(config["operator_address_1"])
    appointee_addr = Web3.to_checksum_address(config["operator_address_1"])
    targets, selectors = clients.el_reader.list_appointee_permissions(account_addr, appointee_addr)
    assert len(targets) == len(selectors)
    for target in targets:
        assert Web3.is_address(target)
    for selector in selectors:
        assert isinstance(selector, bytes)
    print(
        f"Appointee permissions: targets={targets}, selectors=[{', '.join(['0x' + s.hex() for s in selectors])}]"
    )


def test_list_pending_admins():
    account_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.list_pending_admins(account_addr)
    assert isinstance(result, list)
    for admin in result:
        assert Web3.is_address(admin)
    print(f"Pending admins: {result}")


def test_list_admins():
    account_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.list_admins(account_addr)
    assert isinstance(result, list)
    for admin in result:
        assert Web3.is_address(admin)
    print(f"Admins: {result}")


def test_is_pending_admin():
    account_addr = Web3.to_checksum_address(config["operator_address_1"])
    pending_admin_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.is_pending_admin(account_addr, pending_admin_addr)
    assert isinstance(result, bool)
    print(f"Is pending admin: {result}")


def test_is_admin():
    account_addr = Web3.to_checksum_address(config["operator_address_1"])
    admin_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.is_admin(account_addr, admin_addr)
    assert isinstance(result, bool)
    print(f"Is admin: {result}")


def test_get_distribution_roots_length():
    result = clients.el_reader.get_distribution_roots_length()
    assert isinstance(result, int)
    print(f"Distribution roots length: {result}")


def test_get_curr_rewards_calculation_end_timestamp():
    result = clients.el_reader.get_curr_rewards_calculation_end_timestamp()
    assert isinstance(result, int)
    print(f"Current rewards calculation end timestamp: {result}")


def test_get_current_claimable_distribution_root():
    result = clients.el_reader.get_current_claimable_distribution_root()
    assert isinstance(result, dict)
    assert "root" in result
    assert "startBlock" in result
    assert "endBlock" in result
    assert "totalClaimable" in result
    print(f"Current claimable distribution root: {result}")


def test_get_cumulative_claimed():
    earner_addr = Web3.to_checksum_address(config["operator_address_1"])
    token_addr = Web3.to_checksum_address(config["strategy_addr"])
    result = clients.el_reader.get_cumulative_claimed(earner_addr, token_addr)
    assert isinstance(result, int)
    print(f"Cumulative claimed: {result}")


def test_get_operator_avs_split():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    result = clients.el_reader.get_operator_avs_split(operator_addr, avs_addr)
    assert isinstance(result, int)
    print(f"Operator AVS split: {result}")


def test_get_operator_pi_split():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.get_operator_pi_split(operator_addr)
    assert isinstance(result, int)
    print(f"Operator PI split: {result}")


def test_get_operator_set_split():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    operator_sets = clients.el_reader.get_operator_sets_for_operator(operator_addr)
    if operator_sets:
        operator_set = operator_sets[0]
    else:
        operator_set = {
            "Id": 0,
            "Avs": Web3.to_checksum_address(config["avs_registry_coordinator_address"]),
        }
    result = clients.el_reader.get_operator_set_split(operator_addr, operator_set)
    assert isinstance(result, int)
    print(f"Operator set split: {result}")


def test_get_rewards_updater():
    result = clients.el_reader.get_rewards_updater()
    assert Web3.is_address(result)
    print(f"Rewards updater: {result.hex() if isinstance(result, bytes) else result}")


def test_get_default_operator_split_bips():
    result = clients.el_reader.get_default_operator_split_bips()
    assert isinstance(result, int)
    print(f"Default operator split bips: {result}")


def test_get_claimer_for():
    earner_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.get_claimer_for(earner_addr)
    assert Web3.is_address(result)
    print(f"Claimer for {earner_addr}: {result.hex() if isinstance(result, bytes) else result}")


def test_get_submission_nonce():
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    result = clients.el_reader.get_submission_nonce(avs_addr)
    assert isinstance(result, int)
    print(f"Submission nonce: {result}")


def test_get_is_avs_rewards_submission_hash():
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    hash_value = nums_to_bytes([0] * 32)
    result = clients.el_reader.get_is_avs_rewards_submission_hash(avs_addr, hash_value)
    assert isinstance(result, bool)
    print(f"Is AVS rewards submission hash: {result}")


def test_get_is_rewards_submission_for_all_hash():
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    hash_value = nums_to_bytes([0] * 32)
    result = clients.el_reader.get_is_rewards_submission_for_all_hash(avs_addr, hash_value)
    assert isinstance(result, bool)
    print(f"Is rewards submission for all hash: {result}")


def test_get_is_rewards_for_all_submitter():
    submitter_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.get_is_rewards_for_all_submitter(submitter_addr)
    assert isinstance(result, bool)
    print(f"Is rewards for all submitter: {result}")


def test_get_is_rewards_submission_for_all_earners_hash():
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    hash_value = nums_to_bytes([0] * 32)
    result = clients.el_reader.get_is_rewards_submission_for_all_earners_hash(avs_addr, hash_value)
    assert isinstance(result, bool)
    print(f"Is rewards submission for all earners hash: {result}")


def test_get_is_operator_directed_avs_rewards_submission_hash():
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    hash_value = nums_to_bytes([0] * 32)
    result = clients.el_reader.get_is_operator_directed_avs_rewards_submission_hash(
        avs_addr, hash_value
    )
    assert isinstance(result, bool)
    print(f"Is operator directed AVS rewards submission hash: {result}")


def test_get_is_operator_directed_operator_set_rewards_submission_hash():
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    hash_value = nums_to_bytes([0] * 32)
    result = clients.el_reader.get_is_operator_directed_operator_set_rewards_submission_hash(
        avs_addr, hash_value
    )
    assert isinstance(result, bool)
    print(f"Is operator directed operator set rewards submission hash: {result}")


def test_get_strategy_and_underlying_token():
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    result = clients.el_reader.get_strategy_and_underlying_token(strategy_addr)
    assert isinstance(result, tuple)
    assert len(result) == 2
    strategy_contract, token_addr = result
    assert Web3.is_address(token_addr)
    print(
        f"Strategy and underlying token: contract={strategy_contract.address}, token_addr={token_addr}"
    )


def test_get_strategy_and_underlying_erc20_token():
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    result = clients.el_reader.get_strategy_and_underlying_erc20_token(strategy_addr)
    assert isinstance(result, tuple)
    assert len(result) == 3
    strategy_contract, token_contract, token_addr = result
    assert Web3.is_address(token_addr)
    print(
        f"Strategy and underlying ERC20 token: token_addr={token_addr}, token_contract={token_contract.address}, strategy_contract={strategy_contract.address}"
    )


def test_calculate_operator_avs_registration_digest_hash():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    avs_addr = Web3.to_checksum_address(config["avs_address"])
    salt = nums_to_bytes([0] * 32)
    expiry = 2**32 - 1
    result = clients.el_reader.calculate_operator_avs_registration_digest_hash(
        operator_addr, avs_addr, salt, expiry
    )
    assert isinstance(result, bytes)
    assert len(result) == 32
    print(f"Operator AVS registration digest hash: 0x{result.hex()}")


def test_get_encumbered_magnitude():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    result = clients.el_reader.get_encumbered_magnitude(operator_addr, strategy_addr)
    assert isinstance(result, int)
    print(f"Encumbered magnitude: {result}")


def test_get_calculation_interval_seconds():
    result = clients.el_reader.get_calculation_interval_seconds()
    assert isinstance(result, int)
    print(f"Calculation interval seconds: {result}")


def test_get_max_rewards_duration():
    result = clients.el_reader.get_max_rewards_duration()
    assert isinstance(result, int)
    print(f"Max rewards duration: {result}")


def test_get_max_retroactive_length():
    result = clients.el_reader.get_max_retroactive_length()
    assert isinstance(result, int)
    print(f"Max retroactive length: {result}")


def test_get_max_future_length():
    result = clients.el_reader.get_max_future_length()
    assert isinstance(result, int)
    print(f"Max future length: {result}")


def test_get_genesis_rewards_timestamp():
    result = clients.el_reader.get_genesis_rewards_timestamp()
    assert isinstance(result, int)
    print(f"Genesis rewards timestamp: {result}")


def test_get_activation_delay():
    result = clients.el_reader.get_activation_delay()
    assert isinstance(result, int)
    print(f"Activation delay: {result}")


def test_get_deallocation_delay():
    result = clients.el_reader.get_deallocation_delay()
    assert isinstance(result, int)
    print(f"Deallocation delay: {result}")


def test_get_allocation_configuration_delay():
    result = clients.el_reader.get_allocation_configuration_delay()
    assert isinstance(result, int)
    print(f"Allocation configuration delay: {result}")


def test_get_num_operator_sets_for_operator():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    result = clients.el_reader.get_num_operator_sets_for_operator(operator_addr)
    print(f"Number of operator sets for operator: {result}")


def test_get_slashable_shares():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    operator_sets = clients.el_reader.get_operator_sets_for_operator(operator_addr)
    if operator_sets:
        operator_set = operator_sets[0]
    else:
        operator_set = {
            "id": 0,
            "avs": Web3.to_checksum_address(config["avs_registry_coordinator_address"]),
        }
    result = clients.el_reader.get_slashable_shares(operator_addr, operator_set, [strategy_addr])
    print(f"Slashable shares: {result}")


def test_get_slashable_shares_for_operator_sets_before():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    operator_sets = clients.el_reader.get_operator_sets_for_operator(operator_addr)
    if not operator_sets:
        operator_sets = [
            {"id": 1, "avs": Web3.to_checksum_address(config["avs_registry_coordinator_address"])}
        ]
    current_block = clients.eth_http_client.eth.block_number
    result = clients.el_reader.get_slashable_shares_for_operator_sets_before(
        operator_sets, current_block
    )
    print(f"Slashable shares for operator sets before: {result}")


def test_get_slashable_shares_for_operator_sets():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    operator_sets = clients.el_reader.get_operator_sets_for_operator(operator_addr)
    if not operator_sets:
        operator_sets = [
            {"id": 1, "avs": Web3.to_checksum_address(config["avs_registry_coordinator_address"])}
        ]
    result = clients.el_reader.get_slashable_shares_for_operator_sets(operator_sets)
    assert isinstance(result, list)
    for item in result:
        assert isinstance(item, dict)
        assert "OperatorSet" in item
        assert "Strategies" in item
        assert "Operators" in item
        assert "SlashableStakes" in item
        operator_set = item["OperatorSet"]
        assert isinstance(operator_set, dict)
        assert "id" in operator_set
        assert "avs" in operator_set
        assert isinstance(operator_set["id"], int)
        assert Web3.is_address(operator_set["avs"])
        assert isinstance(item["Strategies"], list)
        assert isinstance(item["Operators"], list)
        assert isinstance(item["SlashableStakes"], list)
        for strategy in item["Strategies"]:
            assert Web3.is_address(strategy)
        for operator in item["Operators"]:
            assert Web3.is_address(operator)
        for stake in item["SlashableStakes"]:
            assert isinstance(stake, int)
    print(f"Slashable shares for operator sets: {result}")



import ctypes

from . import hook
from . import utils

BUFFER_SIZE = 2048


def buildClear(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_clear",
        None,
        [ctypes.POINTER(cls)],
    )

    def clear(self):
        return wrapper(self)

    return clear


def buildSetStr(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_setStr",
        None,
        [ctypes.POINTER(cls), ctypes.c_char_p, ctypes.c_size_t, ctypes.c_int64],
    )

    def setStr(self, value, mode=10):
        return wrapper(self, value, len(value), mode)

    return setStr


def buildSetInt(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_setInt",
        None,
        [ctypes.POINTER(cls), ctypes.c_int64],
    )

    def setInt(self, value):
        return wrapper(self, value)

    return setInt


def buildSetByCSPRNG(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_setByCSPRNG",
        None,
        [ctypes.POINTER(cls)],
    )

    def setByCSPRNG(self):
        return wrapper(self)

    return setByCSPRNG


def buildNormalize(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_normalize",
        None,
        [ctypes.POINTER(cls), ctypes.POINTER(cls)],
    )

    def normalize(self):
        result = cls()
        wrapper(result, self)
        return result

    return normalize


def buildGetSubArray(cls, idx, sub_cls):
    def getSubArray(self):
        addr = ctypes.addressof(self)
        result = sub_cls.from_address(addr + (ctypes.sizeof(sub_cls) * idx))
        return result

    return getSubArray


def buildGetStr(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_getStr",
        None,
        [
            (ctypes.c_char * (BUFFER_SIZE + 1)),
            ctypes.c_size_t,
            ctypes.POINTER(cls),
            ctypes.c_uint64,
        ],
    )

    def getStr(self, mode=10):
        buffer = ctypes.create_string_buffer(b"\0" * BUFFER_SIZE)
        wrapper(buffer, BUFFER_SIZE, self, mode)
        return buffer.value

    return getStr


def buildIsEqual(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_isEqual",
        ctypes.c_int64,
        [ctypes.POINTER(cls), ctypes.POINTER(cls)],
    )

    def isEqual(self, other):
        return wrapper(self, other) == 1

    return isEqual


def buildIsOne(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_isOne",
        None,
        [ctypes.POINTER(cls)],
    )

    def isOne(self, other):
        return wrapper(self) == 1

    return isOne


def buildIsZero(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_isZero",
        ctypes.c_int64,
        [ctypes.POINTER(cls)],
    )

    def isZero(self, other):
        return wrapper(self) == 1

    return isZero


def buildThreeOp(cls, op_name):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_{op_name}",
        None,
        [ctypes.POINTER(cls), ctypes.POINTER(cls), ctypes.POINTER(cls)],
    )

    def op(self, other):
        result = cls()
        wrapper(result, self, other)
        return result

    op.__name__ = op_name
    return op


def buildTwoOp(cls, op_name):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_{op_name}",
        None,
        [ctypes.POINTER(cls), ctypes.POINTER(cls)],
    )

    def op(self):
        result = cls()
        wrapper(result, self)
        return result

    op.__name__ = op_name
    return op


def buildMul(cls, right_op):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_mul",
        None,
        [ctypes.POINTER(cls), ctypes.POINTER(cls), ctypes.POINTER(right_op)],
    )

    def mul(self, right):
        result = cls()
        wrapper(result, self, right)
        return result

    return mul


def buildSerialize(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_serialize",
        ctypes.c_size_t,
        [
            (ctypes.c_char * (BUFFER_SIZE + 1)),
            ctypes.c_size_t,
            ctypes.POINTER(cls),
            ctypes.c_uint64,
        ],
    )

    def serialize(self, mode=10):
        buffer = ctypes.create_string_buffer(b"\0" * BUFFER_SIZE)
        len = wrapper(buffer, BUFFER_SIZE, self, mode)
        return bytes(bytearray(buffer[:len]))

    return serialize


def buildDeserialize(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_deserialize",
        None,
        [
            ctypes.POINTER(cls),
            (ctypes.c_char * (BUFFER_SIZE + 1)),
            ctypes.c_size_t,
        ],
    )

    def deserialize(self, value):
        if len(value) > BUFFER_SIZE:
            raise RuntimeError("deserialize error: value is more than a buffer size.")
        buffer = ctypes.create_string_buffer(b"\0" * BUFFER_SIZE)
        ctypes.memmove(buffer, value, len(value))
        wrapper(self, buffer, BUFFER_SIZE)

    return deserialize


def buildSetHashOf(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_setHashOf",
        None,
        [ctypes.POINTER(cls), ctypes.c_char_p, ctypes.c_size_t],
    )

    def setHashOf(self, value):
        return wrapper(self, ctypes.c_char_p(value), len(value))

    return setHashOf


def buildHashAndMapTo(cls):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        f"mclBn{cls.__name__}_hashAndMapTo",
        None,
        [ctypes.POINTER(cls), ctypes.c_char_p, ctypes.c_size_t],
    )

    @staticmethod
    def hashAndMapTo(value):
        result = cls()
        wrapper(result, ctypes.c_char_p(value), len(value))
        return result

    return hashAndMapTo


def buildPairing(cls, left_group, right_group):
    wrapper = utils.wrap_function(
        hook.mclbn384_256,
        "mclBn_pairing",
        None,
        [
            ctypes.POINTER(cls),
            ctypes.POINTER(left_group),
            ctypes.POINTER(right_group),
        ],
    )

    @staticmethod
    def pairing(g1, g2):
        result = cls()
        wrapper(result, g1, g2)
        # print(f"paring: g1: {g1.serialize().hex()}")
        # print(f"paring: g2: {g2.getStr()}")
        # print(f"paring: gt: {result.getStr()}")
        # print(f"paring: done ==============================")
        return result

    return pairing



import ctypes

from .. import builder
from .Fp import Fp
from .Fr import Fr


class G1(ctypes.Structure):
    _fields_ = [
        ("x", Fp),
        ("y", Fp),
        ("z", Fp),
    ]


G1.__add__ = builder.buildThreeOp(G1, "add")
G1.__eq__ = builder.buildIsEqual(G1)
G1.__mul__ = builder.buildMul(G1, Fr)
G1.__neg__ = builder.buildTwoOp(G1, "neg")
G1.__sub__ = builder.buildThreeOp(G1, "sub")
G1.normalize = builder.buildNormalize(G1)
G1.deserialize = builder.buildDeserialize(G1)
G1.getStr = builder.buildGetStr(G1)
G1.hashAndMapTo = builder.buildHashAndMapTo(G1)
G1.isZero = builder.buildIsZero(G1)
G1.serialize = builder.buildSerialize(G1)
G1.setStr = builder.buildSetStr(G1)
G1.getX = builder.buildGetSubArray(G1, 0, Fp)
G1.getY = builder.buildGetSubArray(G1, 1, Fp)
G1.getZ = builder.buildGetSubArray(G1, 2, Fp)
G1.clear = builder.buildClear(G1)



from typing import List, Tuple, Dict, Any

from eth_abi import encode
from eth_account.signers.local import LocalAccount
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.types import ChecksumAddress
from web3.types import TxReceipt

from eigensdk.contracts import ABIs
from eigensdk.crypto.bls.attestation import G1Point, KeyPair


def nums_to_bytes(nums: List[int]) -> bytes:
    return "".join(map(chr, nums)).encode()


def bitmap_to_quorum_ids(bitmap: int) -> List[int]:
    quorum_ids = []
    for i in range(256):
        if bitmap & (1 << i):
            quorum_ids.append(int(i))
    return quorum_ids


def _send_transaction(
    func: ContractFunction,
    pk_wallet: LocalAccount,
    eth_http_client: Web3,
    gas_limit: int = 10_000_000,
    skip_estimation: bool = True,
) -> TxReceipt:
    if skip_estimation:
        gas_estimate = gas_limit
    else:
        try:
            gas_estimate = func.estimate_gas({"from": pk_wallet.address})
        except Exception as e:
            raise Exception(
                f"""Gas estimation failed: {e}. Consider using
                skip_estimation=True with a manual gas_limit."""
            )

    current_gas_price = eth_http_client.eth.gas_price

    tx = func.build_transaction(
        {
            "from": pk_wallet.address,
            "gas": gas_estimate,
            "gasPrice": current_gas_price,
            "nonce": eth_http_client.eth.get_transaction_count(pk_wallet.address),
            "chainId": eth_http_client.eth.chain_id,
        }
    )
    signed_tx = eth_http_client.eth.account.sign_transaction(tx, private_key=pk_wallet.key)
    tx_hash = eth_http_client.eth.send_raw_transaction(signed_tx.raw_transaction)
    return eth_http_client.eth.wait_for_transaction_receipt(tx_hash)


class Transactor:
    def __init__(
        self,
        pk_wallet: LocalAccount,
        eth_http_client: Web3,
        gas_limit: int = 10_000_000,
        skip_estimation: bool = True,
    ):
        self.pk_wallet = pk_wallet
        self.eth_http_client = eth_http_client
        self.gas_limit = gas_limit
        self.skip_estimation = skip_estimation

    def send(self, func: ContractFunction):
        return _send_transaction(
            func, self.pk_wallet, self.eth_http_client, self.gas_limit, self.skip_estimation
        )


def abi_encode_normal_registration_params(
    registration_type: int,
    socket: str,
    pubkey_reg_params: Dict[str, Tuple[int, int] | Tuple[Tuple[int, int], Tuple[int, int]]],
) -> bytes:
    """
    ABI encode a normal registration tuple:
    (uint8, string, ((uint256, uint256), (uint256, uint256), (uint256[2], uint256[2])))
    """

    # Extract structured values
    pubkey_registration_signature = pubkey_reg_params["pubkeyRegistrationSignature"]  # (x, y)
    pubkey_g1 = pubkey_reg_params["pubkeyG1"]  # (x, y)
    pubkey_g2 = pubkey_reg_params["pubkeyG2"]  # ((x0, x1), (y0, y1))

    # Structure expected by ABI
    pubkey_struct = (
        pubkey_registration_signature,  # (uint256, uint256)
        pubkey_g1,  # (uint256, uint256)
        pubkey_g2,  # ((uint256[2], uint256[2]))
    )

    # Full data
    registration_struct = (
        registration_type,  # uint8
        socket,  # string
        pubkey_struct,  # pubkeyRegParams struct
    )

    abi_type = "(uint8,string,((uint256,uint256),(uint256,uint256),(uint256[2],uint256[2])))"

    encoded = encode([abi_type], [registration_struct])

    return encoded[32:]


def get_pubkey_registration_params(
    eth_client: Web3,
    registry_coordinator_addr: ChecksumAddress,
    operator_address: ChecksumAddress,
    bls_key_pair: KeyPair,
) -> Dict[str, Any]:
    # Create contract instance for registry coordinator
    registry_coordinator = eth_client.eth.contract(
        address=registry_coordinator_addr,
        abi=ABIs.REGISTRY_COORDINATOR_ABI,
    )

    g1_hashed_msg_to_sign = registry_coordinator.functions.pubkeyRegistrationMessageHash(
        operator_address
    ).call()

    g1_hashed_msg_as_point = G1Point(*g1_hashed_msg_to_sign)
    signed_msg = bls_key_pair.sign_hashed_to_curve_message(g1_hashed_msg_as_point)

    # Convert public keys to BN254 format
    pubkey_reg_params = {
        "pubkeyRegistrationSignature": (
            int(signed_msg.getX().getStr()),
            int(signed_msg.getY().getStr()),
        ),
        "pubkeyG1": (
            int(bls_key_pair.pub_g1.getX().getStr()),
            int(bls_key_pair.pub_g1.getY().getStr()),
        ),
        "pubkeyG2": (
            (
                int(bls_key_pair.pub_g2.getX().get_a().getStr()),
                int(bls_key_pair.pub_g2.getX().get_b().getStr()),
            ),
            (
                int(bls_key_pair.pub_g2.getY().get_a().getStr()),
                int(bls_key_pair.pub_g2.getY().get_b().getStr()),
            ),
        ),
    }

    print(pubkey_reg_params)
    return pubkey_reg_params



from dataclasses import dataclass
from typing import Any

from web3 import Web3
from web3.types import ChecksumAddress

from eigensdk.crypto.bls.attestation import G1Point, G2Point, Signature


@dataclass
class Operator:
    address: ChecksumAddress
    earnings_receiver_address: ChecksumAddress
    staker_opt_out_window_blocks: int
    metadata_url: str
    allocation_delay: int = 50
    delegation_approver_address: ChecksumAddress = Web3.to_checksum_address(
        "0x0000000000000000000000000000000000000000"
    )


@dataclass
class OperatorPubkeys:
    g1_pub_key: G1Point
    g2_pub_key: G2Point


@dataclass
class OperatorInfo:
    socket: str
    pub_keys: OperatorPubkeys


@dataclass
class OperatorAvsState:
    operator_id: bytes
    operator_info: OperatorInfo
    stake_per_quorum: dict[int, int]
    block_number: int


@dataclass
class QuorumAvsState:
    quorum_number: int
    total_stake: int
    agg_pub_key_g1: G1Point
    block_number: int


@dataclass
class OperatorStateRetrieverCheckSignaturesIndices:
    non_signer_quorum_bitmap_indices: list[int]
    quorum_apk_indices: list[int]
    total_stake_indices: list[int]
    non_signer_stake_indices: list[list[int]]


@dataclass
class SignedTaskResponseDigest:
    task_response: Any
    bls_signature: Signature
    operator_id: int


@dataclass
class OperatorStateRetrieverOperator:
    operator: ChecksumAddress
    operator_id: bytes
    stake: int


@dataclass
class StakeRegistryTypesStrategyParams:
    strategy: str
    multiplier: int


@dataclass
class StakeRegistryTypesStakeUpdate:
    update_block_number: int
    next_update_block_number: int
    stake: int


@dataclass
class BLSApkRegistryTypesApkUpdate:
    apk_hash: bytes
    update_block_number: int
    next_update_block_number: int



# Configuration file for the Sphinx documentation builder.

# -- Project information

project = "EigenSDK Python"
copyright = "2025, Abram Symons"
author = "Abram Symons"

release = "0.1.0"
version = "0.1.0"

# -- General configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# -- Options for HTML output

html_theme = "sphinx_rtd_theme"

# -- Options for EPUB output

epub_show_urls = "footnote"

# -- Custom project attribution note (can be used in templates or footers)

rst_epilog = """
.. |project_note| replace:: This SDK was originally developed by Abram Symons .
"""



import ctypes

from .. import builder
from .Fp2 import Fp2
from .Fr import Fr


class G2(ctypes.Structure):
    _fields_ = [
        ("x", Fp2),
        ("y", Fp2),
        ("z", Fp2),
    ]


G2.__add__ = builder.buildThreeOp(G2, "add")
G2.__eq__ = builder.buildIsEqual(G2)
G2.__mul__ = builder.buildMul(G2, Fr)
G2.__neg__ = builder.buildTwoOp(G2, "neg")
G2.__sub__ = builder.buildThreeOp(G2, "sub")
G2.normalize = builder.buildNormalize(G2)
G2.deserialize = builder.buildDeserialize(G2)
G2.getStr = builder.buildGetStr(G2)
G2.hashAndMapTo = builder.buildHashAndMapTo(G2)
G2.isZero = builder.buildIsZero(G2)
G2.serialize = builder.buildSerialize(G2)
G2.setStr = builder.buildSetStr(G2)
G2.getX = builder.buildGetSubArray(G2, 0, Fp2)
G2.getY = builder.buildGetSubArray(G2, 1, Fp2)
G2.getZ = builder.buildGetSubArray(G2, 2, Fp2)
G2.clear = builder.buildClear(G2)



import logging
from typing import Tuple
from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_typing import Address
from web3 import Web3
from eigensdk.chainio.clients.avsregistry import reader as avs_reader
from eigensdk.chainio.clients.avsregistry import writer as avs_writer
from eigensdk.chainio.clients.elcontracts import reader as el_reader
from eigensdk.chainio.clients.elcontracts import writer as el_writer
from eigensdk.contracts import ABIs


class BuildAllConfig:
    def __init__(
        self,
        eth_http_url: str,
        registry_coordinator_addr: Address,
        operator_state_retriever_addr: Address,
        rewards_coordinator_addr: Address,
        permission_controller_addr: Address,
        service_manager_addr: Address,
        allocation_manager_addr: Address,
        delegation_manager_addr: Address,
        avs_name: str,
    ) -> None:

        self.eth_http_url: str = eth_http_url
        self.registry_coordinator_addr = Web3.to_checksum_address(registry_coordinator_addr)
        self.operator_state_retriever_addr = Web3.to_checksum_address(operator_state_retriever_addr)
        self.rewards_coordinator_addr = Web3.to_checksum_address(rewards_coordinator_addr)
        self.permission_controller_addr = Web3.to_checksum_address(permission_controller_addr)
        self.service_manager_addr = Web3.to_checksum_address(service_manager_addr)
        self.allocation_manager_addr = Web3.to_checksum_address(allocation_manager_addr)
        self.delegation_manager_addr = Web3.to_checksum_address(delegation_manager_addr)
        self.avs_name: str = avs_name
        self.logger: logging.Logger = logging.getLogger(__name__)

    def build_el_clients(
        self,
        ecdsa_private_key: str,
    ) -> Tuple[el_reader.ELReader, el_writer.ELWriter]:
        eth_http_client = Web3(Web3.HTTPProvider(self.eth_http_url))

        pk_wallet = Account.from_key(ecdsa_private_key)
        registry_coordinator_instance = eth_http_client.eth.contract(
            address=self.registry_coordinator_addr,
            abi=ABIs.REGISTRY_COORDINATOR_ABI,
        )
        self.logger.info(f"registry_coordinator_instance: {registry_coordinator_instance}")
        stake_registry_addr = registry_coordinator_instance.functions.stakeRegistry().call()
        stake_registry_instance = eth_http_client.eth.contract(
            address=stake_registry_addr,
            abi=ABIs.STAKE_REGISTRY_ABI,
        )
        self.logger.info(f"stake_registry_instance: {stake_registry_instance}")
        delegation_manager_addr = stake_registry_instance.functions.delegation().call()
        delegation_manager_instance = eth_http_client.eth.contract(
            address=delegation_manager_addr,
            abi=ABIs.DELEGATION_MANAGER_ABI,
        )
        self.logger.info(f"delegation_manager_instance: {delegation_manager_instance}")
        strategy_manager_addr = delegation_manager_instance.functions.strategyManager().call()
        strategy_manager_instance = eth_http_client.eth.contract(
            address=strategy_manager_addr,
            abi=ABIs.STRATEGY_MANAGER_ABI,
        )
        self.logger.info(f"strategy_manager_instance: {strategy_manager_instance}")
        service_manager = eth_http_client.eth.contract(
            address=self.service_manager_addr,
            abi=ABIs.SERVICE_MANAGER_BASE_ABI,
        )

        allocation_manager_instance = eth_http_client.eth.contract(
            address=self.allocation_manager_addr,
            abi=ABIs.ALLOCATION_MANAGER_ABI,
        )
        self.logger.info(f"allocation_manager_instance: {allocation_manager_instance}")
        permission_controller_instance = eth_http_client.eth.contract(
            address=self.permission_controller_addr,
            abi=ABIs.PERMISSION_CONTROLLER_ABI,
        )
        self.logger.info(f"permission_controller_instance: {permission_controller_instance}")
        avs_directory_addr = service_manager.functions.avsDirectory().call()
        avs_directory_instance = eth_http_client.eth.contract(
            address=avs_directory_addr,
            abi=ABIs.AVS_DIRECTORY_ABI,
        )
        self.logger.info(f"avs_directory_instance: {avs_directory_instance}")
        rewards_coordinator_instance = eth_http_client.eth.contract(
            address=self.rewards_coordinator_addr, abi=ABIs.REWARDS_COORDINATOR_ABI
        )
        self.logger.info(f"rewards_coordinator_instance: {rewards_coordinator_instance}")
        el_reader_instance = el_reader.ELReader(
            allocation_manager=allocation_manager_instance,
            avs_directory=avs_directory_instance,
            delegation_manager=delegation_manager_instance,
            permission_controller=permission_controller_instance,
            reward_coordinator=rewards_coordinator_instance,
            strategy_manager=strategy_manager_instance,
            logger=self.logger,
            eth_http_client=eth_http_client,
            strategy_abi=ABIs.I_STRATEGY_ABI,
            erc20_abi=ABIs.IERC20_ABI,
        )
        el_writer_instance = el_writer.ELWriter(
            allocation_manager=allocation_manager_instance,
            avs_directory=avs_directory_instance,
            delegation_manager=delegation_manager_instance,
            permission_controller=permission_controller_instance,
            reward_coordinator=rewards_coordinator_instance,
            registry_coordinator=registry_coordinator_instance,
            strategy_manager=strategy_manager_instance,
            el_chain_reader=el_reader_instance,
            logger=self.logger,
            pk_wallet=pk_wallet,
            eth_http_client=eth_http_client,
            strategy_abi=ABIs.I_STRATEGY_ABI,
            erc20_abi=ABIs.IERC20_ABI,
        )

        return el_reader_instance, el_writer_instance

    def build_avs_registry_clients(
        self,
        ecdsa_private_key: str,
        el_chain_reader: el_reader.ELReader,
    ) -> Tuple[avs_reader.AvsRegistryReader, avs_writer.AvsRegistryWriter]:
        pk_wallet = Account.from_key(ecdsa_private_key)
        eth_http_client = Web3(Web3.HTTPProvider(self.eth_http_url))
        registry_coordinator_instance = eth_http_client.eth.contract(
            address=self.registry_coordinator_addr,
            abi=ABIs.REGISTRY_COORDINATOR_ABI,
        )
        self.logger.info(f"registry_coordinator_instance: {registry_coordinator_instance}")
        operator_state_retriever_instance = eth_http_client.eth.contract(
            address=self.operator_state_retriever_addr, abi=ABIs.OPERATOR_STATE_RETRIEVER_ABI
        )
        self.logger.info(f"operator_state_retriever_instance: {operator_state_retriever_instance}")
        bls_apk_registry_addr = registry_coordinator_instance.functions.blsApkRegistry().call()
        bls_apk_registry_instance = eth_http_client.eth.contract(
            address=bls_apk_registry_addr,
            abi=ABIs.BLS_APK_REGISTRY_ABI,
        )
        self.logger.info(f"bls_apk_registry_instance: {bls_apk_registry_instance}")
        service_manager_instance = eth_http_client.eth.contract(
            address=self.service_manager_addr,
            abi=ABIs.SERVICE_MANAGER_BASE_ABI,
        )
        self.logger.info(f"service_manager_instance: {service_manager_instance}")
        stake_registry_addr = registry_coordinator_instance.functions.stakeRegistry().call()
        stake_registry_instance = eth_http_client.eth.contract(
            address=stake_registry_addr,
            abi=ABIs.STAKE_REGISTRY_ABI,
        )
        self.logger.info(f"stake_registry_instance: {stake_registry_instance}")
        avs_reader_instance = avs_reader.AvsRegistryReader(
            registry_coordinator=registry_coordinator_instance,
            registry_coordinator_addr=Address(bytes.fromhex(self.registry_coordinator_addr[2:])),
            bls_apk_registry=bls_apk_registry_instance,
            bls_apk_registry_addr=bls_apk_registry_addr,
            operator_state_retriever=operator_state_retriever_instance,
            service_manager=service_manager_instance,
            stake_registry=stake_registry_instance,
            logger=self.logger,
            eth_http_client=eth_http_client,
        )
        avs_writer_instance = avs_writer.AvsRegistryWriter(
            registry_coordinator=registry_coordinator_instance,
            operator_state_retriever=operator_state_retriever_instance,
            service_manager=service_manager_instance,
            service_manager_addr=Address(bytes.fromhex(self.service_manager_addr[2:])),
            stake_registry=stake_registry_instance,
            bls_apk_registry=bls_apk_registry_instance,
            el_reader=el_chain_reader,
            logger=self.logger,
            eth_http_client=eth_http_client,
            pk_wallet=pk_wallet,
        )
        return avs_reader_instance, avs_writer_instance


class Clients:
    def __init__(
        self,
        avs_registry_reader: avs_reader.AvsRegistryReader,
        avs_registry_writer: avs_writer.AvsRegistryWriter,
        el_reader: el_reader.ELReader,
        el_writer: el_writer.ELWriter,
        eth_http_client: Web3,
        wallet: LocalAccount,
    ):
        self.avs_registry_reader = avs_registry_reader
        self.avs_registry_writer = avs_registry_writer
        self.el_reader = el_reader
        self.el_writer = el_writer
        self.eth_http_client = eth_http_client
        self.wallet = wallet


def build_all(
    config: BuildAllConfig,
    config_ecdsa_private_key: str,
) -> Clients:
    eth_http_client = Web3(Web3.HTTPProvider(config.eth_http_url))
    pk_wallet: LocalAccount = Account.from_key(config_ecdsa_private_key)

    el_reader, el_writer = config.build_el_clients(ecdsa_private_key=config_ecdsa_private_key)

    avs_reader, avs_writer = config.build_avs_registry_clients(
        ecdsa_private_key=config_ecdsa_private_key, el_chain_reader=el_reader
    )

    return Clients(
        avs_registry_reader=avs_reader,
        avs_registry_writer=avs_writer,
        el_reader=el_reader,
        el_writer=el_writer,
        eth_http_client=eth_http_client,
        wallet=pk_wallet,
    )



import os
import yaml  # type: ignore
from eigensdk.chainio.clients.builder import BuildAllConfig, build_all

# Get the directory where builder.py is located
current_dir = os.path.dirname(os.path.abspath(__file__))
config_path = os.path.join(current_dir, "config", "anvil.yaml")

with open(config_path, "r") as f:
    config = yaml.load(f, Loader=yaml.BaseLoader)


cfg = BuildAllConfig(
    eth_http_url=config["eth_rpc_url"],
    avs_name="incredible-squaring",
    registry_coordinator_addr=config["avs_registry_coordinator_address"],
    operator_state_retriever_addr=config["operator_state_retriever_address"],
    rewards_coordinator_addr=config["rewards_coordinator_address"],
    permission_controller_addr=config["permission_controller_address"],
    service_manager_addr=config["service_manager_address"],
    allocation_manager_addr=config["allocation_manager_address"],
    delegation_manager_addr=config["delegation_manager_address"],
)
clients_array = []
for i in range(3):
    clients_array.append(build_all(cfg, config["ecdsa_private_key_{}".format(i + 1)]))

clients = clients_array[0]



import contextlib
import ctypes
import os
import platform
import sys
import textwrap

from . import consts

DIR_FOR_LINKER = os.environ.get("MCL_PATH", "/usr/local")


@contextlib.contextmanager
def change_cwd(path):
    current_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(current_cwd)


def get_dll(path, *args):
    try:
        return ctypes.CDLL(path, *args)
    except OSError:
        print(
            textwrap.dedent(
                f"""
        Failed to import mcl shared library from:

        {DIR_FOR_LINKER}

        Please set your mcl installation dir path to MCL_PATH and run again

        export MCL_PATH=<path to mcl library>

        """
            )
        )
        sys.exit(1)


with change_cwd(DIR_FOR_LINKER):
    system = platform.system()
    if system == "Darwin":
        mclbn384_256 = get_dll("lib/libmclbn384_256.dylib")
    elif system == "Linux":
        get_dll("lib/libmcl.so", ctypes.RTLD_GLOBAL)
        mclbn384_256 = get_dll("lib/libmclbn384_256.so")
    else:
        raise RuntimeError(f"Unsupported OS {system}")


# ret = mclbn384_256.mclBn_init(consts.BN384_256, consts.MCLBN_COMPILED_TIME_VAR)
ret = mclbn384_256.mclBn_init(consts.BN_SNARK1, consts.MCLBN_COMPILED_TIME_VAR)

if ret:
    raise RuntimeError(f"mclbn384_256 ret {ret}")



def wrap_function(lib, funcname, restype, argtypes):
    """Simplify wrapping ctypes functions"""
    func = lib.__getattr__(funcname)
    func.restype = restype
    func.argtypes = argtypes
    return func



from web3 import Web3
from tests.builder import clients_array, clients, config
from eigensdk.crypto.bls import KeyPair
from eth_utils import to_checksum_address
from eigensdk.chainio.clients.elcontracts.writer import Operator


def test_register_as_operator_for_operator_sets():
    for i in range(1, 4):
        address = to_checksum_address(config["operator_address_{}".format(i)])
        operator = Operator(
            address=address,
            earnings_receiver_address=address,
            delegation_approver_address=Web3.to_checksum_address(
                "0x0000000000000000000000000000000000000000"
            ),
            allocation_delay=50,
            metadata_url="https://example.com/operator-metadata",
            staker_opt_out_window_blocks=100,
        )
        receipt = clients_array[i - 1].el_writer.register_as_operator(operator)
        assert receipt is not None
        assert receipt["status"] == 1
        print(f"Registered operator with tx hash: {receipt['transactionHash'].hex()}")
        request = {
            "operator_address": address,
            "avs_address": config["service_manager_address"],
            "operator_set_ids": [0],
            "socket": "operator-socket",
            "bls_key_pair": KeyPair(),
        }
        receipt = clients_array[i - 1].el_writer.register_for_operator_sets(
            config["avs_registry_coordinator_address"], request
        )
        assert receipt["status"] == 1
        print(f"Registered for operator sets with tx hash: {receipt['transactionHash'].hex()}")


def test_update_stakes_of_entire_operator_set_for_quorums():
    operator_addr = Web3.to_checksum_address(config["operator_address_2"])
    operators_per_quorum = [[str(operator_addr)]]
    quorum_numbers = [0]
    receipt = clients.avs_registry_writer.update_stakes_of_entire_operator_set_for_quorums(
        operators_per_quorum, quorum_numbers
    )
    assert receipt is not None
    print(f"Updated stakes with tx hash: {receipt['transactionHash'].hex()}")


def test_update_socket():
    new_socket = "192.168.1.100:9000"
    receipt = clients.avs_registry_writer.update_socket(new_socket)
    assert receipt is not None
    print(f"Updated socket with tx hash: {receipt['transactionHash'].hex()}")


def test_set_avs():
    avs_address = Web3.to_checksum_address(config["service_manager_address"])
    receipt = clients.avs_registry_writer.set_avs(str(avs_address))
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set AVS with tx hash: {receipt['transactionHash'].hex()}")


def test_update_stakes_of_operator_subset_for_all_quorums():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    operators = [str(operator_addr)]
    receipt = clients.avs_registry_writer.update_stakes_of_operator_subset_for_all_quorums(
        operators
    )
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Updated stakes for operator subset with tx hash: {receipt['transactionHash'].hex()}")


def test_set_rewards_initiator():
    rewards_initiator_addr = Web3.to_checksum_address(config["operator_address_1"])
    receipt = clients.avs_registry_writer.set_rewards_initiator(str(rewards_initiator_addr))
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set rewards initiator with tx hash: {receipt['transactionHash'].hex()}")


def test_set_minimum_stake_for_quorum():
    quorum_number = 0
    minimum_stake = 1000000
    receipt = clients.avs_registry_writer.set_minimum_stake_for_quorum(quorum_number, minimum_stake)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set minimum stake for quorum with tx hash: {receipt['transactionHash'].hex()}")


def test_create_total_delegated_stake_quorum():

    operator_set_param_dict = {
        "MaxOperatorCount": 10,
        "KickBIPsOfOperatorStake": 10000,
        "KickBIPsOfTotalStake": 2000,
    }

    operator_set_params = (
        operator_set_param_dict["MaxOperatorCount"],
        operator_set_param_dict["KickBIPsOfOperatorStake"],
        operator_set_param_dict["KickBIPsOfTotalStake"],
    )

    minimum_stake_required = 1000000
    strategy_addr = Web3.to_checksum_address(config["strategy_addr"])
    strategy_params = [(str(strategy_addr), 10000)]
    receipt = clients.avs_registry_writer.create_total_delegated_stake_quorum(
        operator_set_params, minimum_stake_required, strategy_params
    )
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Created total delegated stake quorum with tx hash: {receipt['transactionHash'].hex()}")


def test_set_operator_set_params():
    quorum_number = 0
    operator_set_params = {
        "maxOperatorCount": 10,
        "kickBIPsOfOperatorStake": 10000,
        "kickBIPsOfTotalStake": 2000,
    }
    receipt = clients.avs_registry_writer.set_operator_set_params(
        quorum_number, operator_set_params
    )
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set operator set params with tx hash: {receipt['transactionHash'].hex()}")


def test_set_churn_approver():
    churn_approver_address = Web3.to_checksum_address(config["operator_address_1"])
    receipt = clients.avs_registry_writer.set_churn_approver(str(churn_approver_address))
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set churn approver with tx hash: {receipt['transactionHash'].hex()}")


def test_set_ejector():
    ejector_address = Web3.to_checksum_address(config["operator_address_1"])
    receipt = clients.avs_registry_writer.set_ejector(str(ejector_address))
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set ejector with tx hash: {receipt['transactionHash'].hex()}")


def test_modify_strategy_params():
    quorum_number = 0
    strategy_indices = [0]
    multipliers = [8000]
    receipt = clients.avs_registry_writer.modify_strategy_params(
        quorum_number, strategy_indices, multipliers
    )
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Modified strategy parameters with tx hash: {receipt['transactionHash'].hex()}")


def test_set_ejection_cooldown():
    ejection_cooldown = 100
    receipt = clients.avs_registry_writer.set_ejection_cooldown(ejection_cooldown)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Set ejection cooldown with tx hash: {receipt['transactionHash'].hex()}")


def test_update_avs_metadata_uri():
    metadata_uri = "https://example.com/avs-metadata-updated"
    receipt = clients.avs_registry_writer.update_avs_metadata_uri(metadata_uri)
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Updated AVS metadata URI with tx hash: {receipt['transactionHash'].hex()}")


def test_create_avs_rewards_submission():
    strategies = clients.avs_registry_reader.strategy_params_by_index(quorum_number=0, index=0)
    assert strategies is not None, "Strategy params should not be None"
    strategy_addr = Web3.to_checksum_address(strategies.strategy)
    duration = clients.el_reader.get_calculation_interval_seconds()
    token = clients.el_reader.get_strategy_and_underlying_token(strategy_addr)[1]
    assert token is not None, "Token should not be None"
    strategy_and_multiplier = {"strategy": Web3.to_checksum_address(strategy_addr), "multiplier": 1}
    latest_block = clients.eth_http_client.eth.get_block("latest")
    block_time = latest_block["timestamp"]
    start_timestamp = ((block_time // duration) + 1) * duration
    rewards_submission = {
        "strategiesAndMultipliers": [strategy_and_multiplier],
        "token": Web3.to_checksum_address(token),
        "amount": 1000,
        "startTimestamp": start_timestamp,
        "duration": duration,
    }
    receipt = clients.avs_registry_writer.create_avs_rewards_submission([rewards_submission])
    assert receipt is not None
    assert receipt["status"] == 1
    print(f"Created AVS rewards submission with tx hash: {receipt['transactionHash'].hex()}")


def test_create_slashable_stake_quorum():
    operator_set_params = (
        10,  # MaxOperatorCount
        10000,  # KickBIPsOfOperatorStake
        2000,  # KickBIPsOfTotalStake
    )
    minimum_stake_required = 0
    strategies = clients.avs_registry_reader.strategy_params_by_index(quorum_number=0, index=0)
    assert strategies is not None, "Strategy params should not be None"
    strategy_addr = Web3.to_checksum_address(strategies.strategy)
    strategy_param = (str(strategy_addr), 10000)
    look_ahead_period = 50400  # ~1 week in blocks (assuming ~12s block time)
    receipt = clients.avs_registry_writer.create_slashable_stake_quorum(
        operator_set_params, minimum_stake_required, [strategy_param], look_ahead_period
    )
    assert receipt is not None
    # assert receipt["status"] == 1
    print(f"Created slashable stake quorum with tx hash: {receipt['transactionHash'].hex()}")


def test_set_slashable_stake_lookahead():
    quorum_number = 0
    look_ahead_period = 50400  # ~1 week in blocks (assuming ~12s block time)
    receipt = clients.avs_registry_writer.set_slashable_stake_lookahead(
        quorum_number, look_ahead_period
    )
    assert receipt is not None
    if receipt["status"] == 1:
        print(f"Set slashable stake lookahead with tx hash: {receipt['transactionHash'].hex()}")
    else:
        print("Quorum is not Slashable Stake Quorum")


def test_add_strategies():
    strategies = clients.avs_registry_reader.strategy_params_by_index(quorum_number=0, index=0)
    assert strategies is not None, "Strategy params should not be None"
    strategy_addr = Web3.to_checksum_address(strategies.strategy)
    strategy_params = {"strategy": strategy_addr, "multiplier": 10000}
    receipt = clients.avs_registry_writer.add_strategies(
        quorum_number=0, strategy_params=[strategy_params]
    )
    assert receipt is not None
    # assert receipt["status"] == 1
    print(f"Added strategies with tx hash: {receipt['transactionHash'].hex()}")


def test_eject_operator():
    operator_addr = Web3.to_checksum_address(config["operator_address_1"])
    quorum_numbers = [0]  # Eject from quorum 0
    receipt = clients.avs_registry_writer.eject_operator(
        operator_address=str(operator_addr), quorum_numbers=quorum_numbers
    )
    assert receipt is not None
    # assert receipt["status"] == 1
    print(f"Ejected operator with tx hash: {receipt['transactionHash'].hex()}")


def test_remove_strategies():
    quorum_number = 0
    indices_to_remove = [1]
    receipt = clients.avs_registry_writer.remove_strategies(quorum_number, indices_to_remove)
    assert receipt is not None
    # assert receipt["status"] == 1
    print(f"Removed strategies with tx hash: {receipt['transactionHash'].hex()}")


def test_create_operator_directed_avs_rewards_submission():
    strategies = clients.avs_registry_reader.strategy_params_by_index(quorum_number=0, index=0)
    assert strategies is not None, "Strategy params should not be None"
    strategy_addr = Web3.to_checksum_address(strategies.strategy)
    duration = clients.el_reader.get_calculation_interval_seconds()
    token = clients.el_reader.get_strategy_and_underlying_token(strategy_addr)[1]
    assert token is not None, "Token should not be None"
    latest_block = clients.eth_http_client.eth.get_block("latest")
    block_time = latest_block["timestamp"]
    start_timestamp = ((block_time // duration) + 1) * duration
    strategy_and_multiplier = {"strategy": Web3.to_checksum_address(strategy_addr), "multiplier": 1}
    operator_rewards = [(str(Web3.to_checksum_address(config["operator_address_1"])), 1000)]

    rewards_submission = {
        "strategiesAndMultipliers": [strategy_and_multiplier],
        "token": Web3.to_checksum_address(token),
        "operatorRewards": operator_rewards,
        "startTimestamp": start_timestamp,
        "duration": duration,
        "description": "Some Description",
    }

    receipt = clients.avs_registry_writer.create_operator_directed_avs_rewards_submission(
        [rewards_submission]
    )
    assert receipt is not None
    # assert receipt["status"] == 1
    print(f"Created AVS rewards submission with tx hash: {receipt['transactionHash'].hex()}")
